#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Discover all instances of unittest.TestCase in this directory.
"""
# pylint: disable=file-perms

# Import python libs
from __future__ import absolute_import, print_function

import collections
import os
import sys
import time
import warnings

TESTS_DIR = os.path.dirname(os.path.normpath(os.path.abspath(__file__)))
if os.name == "nt":
    TESTS_DIR = TESTS_DIR.replace("\\", "\\\\")
CODE_DIR = os.path.dirname(TESTS_DIR)

# Let's inject CODE_DIR so salt is importable if not there already
if "" in sys.path:
    sys.path.remove("")
if TESTS_DIR in sys.path:
    sys.path.remove(TESTS_DIR)
if CODE_DIR in sys.path and sys.path[0] != CODE_DIR:
    sys.path.remove(CODE_DIR)
if CODE_DIR not in sys.path:
    sys.path.insert(0, CODE_DIR)
if TESTS_DIR not in sys.path:
    sys.path.insert(1, TESTS_DIR)

try:
    import tests

    if not tests.__file__.startswith(CODE_DIR):
        print("Found tests module not from salt in {}".format(tests.__file__))
        sys.modules.pop("tests")
        module_dir = os.path.dirname(tests.__file__)
        if module_dir in sys.path:
            sys.path.remove(module_dir)
        del tests
except ImportError:
    pass

# Import salt libs
from salt.ext import six  # isort:skip

try:
    from tests.support.paths import TMP, SYS_TMP_DIR, INTEGRATION_TEST_DIR
    from tests.support.paths import CODE_DIR as SALT_ROOT
except ImportError as exc:
    try:
        import tests

        print("Found tests module not from salt in {}".format(tests.__file__))
    except ImportError:
        print("Unable to import salt test module")
        print("PYTHONPATH:", os.environ.get("PYTHONPATH"))
    print("Current sys.path:")
    import pprint

    pprint.pprint(sys.path)
    six.reraise(*sys.exc_info())

from tests.integration import TestDaemon, TestDaemonStartFailed  # isort:skip
from tests.multimaster import MultimasterTestDaemon  # isort:skip
import salt.utils.platform  # isort:skip

if not salt.utils.platform.is_windows():
    import resource

# Import Salt Testing libs
from tests.support.parser import PNUM, print_header  # isort:skip
from tests.support.parser.cover import SaltCoverageTestingParser  # isort:skip

XML_OUTPUT_DIR = os.environ.get(
    "SALT_XML_TEST_REPORTS_DIR", os.path.join(TMP, "xml-test-reports")
)
HTML_OUTPUT_DIR = os.environ.get(
    "SALT_HTML_TEST_REPORTS_DIR", os.path.join(TMP, "html-test-reports")
)

TEST_DIR = os.path.dirname(INTEGRATION_TEST_DIR)
try:
    if SALT_ROOT:
        os.chdir(SALT_ROOT)
except OSError as err:
    print("Failed to change directory to salt's source: {0}".format(err))

# Soft and hard limits on max open filehandles
MAX_OPEN_FILES = {
    "integration": {"soft_limit": 3072, "hard_limit": 4096},
    "unit": {"soft_limit": 1024, "hard_limit": 2048},
}

# Combine info from command line options and test suite directories.  A test
# suite is a python package of test modules relative to the tests directory.
TEST_SUITES_UNORDERED = {
    "unit": {"display_name": "Unit", "path": "unit"},
    "kitchen": {"display_name": "Kitchen", "path": "kitchen"},
    "multimaster": {"display_name": "Multimaster", "path": "multimaster"},
    "module": {"display_name": "Module", "path": "integration/modules"},
    "state": {"display_name": "State", "path": "integration/states"},
    "cli": {"display_name": "CLI", "path": "integration/cli"},
    "client": {"display_name": "Client", "path": "integration/client"},
    "doc": {"display_name": "Documentation", "path": "integration/doc"},
    "ext_pillar": {"display_name": "External Pillar", "path": "integration/pillar"},
    "grains": {"display_name": "Grains", "path": "integration/grains"},
    "shell": {"display_name": "Shell", "path": "integration/shell"},
    "runners": {"display_name": "Runners", "path": "integration/runners"},
    "renderers": {"display_name": "Renderers", "path": "integration/renderers"},
    "returners": {"display_name": "Returners", "path": "integration/returners"},
    "ssh-int": {"display_name": "SSH Integration", "path": "integration/ssh"},
    "spm": {"display_name": "SPM", "path": "integration/spm"},
    "loader": {"display_name": "Loader", "path": "integration/loader"},
    "outputter": {"display_name": "Outputter", "path": "integration/output"},
    "fileserver": {"display_name": "Fileserver", "path": "integration/fileserver"},
    "wheel": {"display_name": "Wheel", "path": "integration/wheel"},
    "api": {"display_name": "NetAPI", "path": "integration/netapi"},
    "cloud_provider": {
        "display_name": "Cloud Provider",
        "path": "integration/cloud/clouds",
    },
    "minion": {"display_name": "Minion", "path": "integration/minion"},
    "reactor": {"display_name": "Reactor", "path": "integration/reactor"},
    "proxy": {"display_name": "Proxy", "path": "integration/proxy"},
    "external_api": {"display_name": "ExternalAPIs", "path": "integration/externalapi"},
    "daemons": {"display_name": "Daemon", "path": "integration/daemons"},
    "sdb": {"display_name": "Sdb", "path": "integration/sdb"},
    "logging": {"display_name": "Logging", "path": "integration/logging"},
}

TEST_SUITES = collections.OrderedDict(
    sorted(TEST_SUITES_UNORDERED.items(), key=lambda x: x[0])
)


class SaltTestsuiteParser(SaltCoverageTestingParser):
    support_docker_execution = True
    support_destructive_tests_selection = True
    support_expensive_tests_selection = True
    source_code_basedir = SALT_ROOT

    def _get_suites(
        self,
        include_unit=False,
        include_cloud_provider=False,
        include_proxy=False,
        include_kitchen=False,
        include_multimaster=False,
    ):
        """
        Return a set of all test suites except unit and cloud provider tests
        unless requested
        """
        suites = set(TEST_SUITES.keys())
        if not include_unit:
            suites -= set(["unit"])
        if not include_cloud_provider:
            suites -= set(["cloud_provider"])
        if not include_proxy:
            suites -= set(["proxy"])
        if not include_kitchen:
            suites -= set(["kitchen"])
        if not include_multimaster:
            suites -= set(["multimaster"])

        return suites

    def _check_enabled_suites(
        self,
        include_unit=False,
        include_cloud_provider=False,
        include_proxy=False,
        include_kitchen=False,
        include_multimaster=False,
    ):
        """
        Query whether test suites have been enabled
        """
        suites = self._get_suites(
            include_unit=include_unit,
            include_cloud_provider=include_cloud_provider,
            include_proxy=include_proxy,
            include_kitchen=include_kitchen,
            include_multimaster=include_multimaster,
        )

        return any([getattr(self.options, suite) for suite in suites])

    def _enable_suites(
        self,
        include_unit=False,
        include_cloud_provider=False,
        include_proxy=False,
        include_kitchen=False,
        include_multimaster=False,
    ):
        """
        Enable test suites for current test run
        """
        suites = self._get_suites(
            include_unit=include_unit,
            include_cloud_provider=include_cloud_provider,
            include_proxy=include_proxy,
            include_kitchen=include_kitchen,
            include_multimaster=include_multimaster,
        )

        for suite in suites:
            setattr(self.options, suite, True)

    def setup_additional_options(self):
        self.add_option(
            "--sysinfo",
            default=False,
            action="store_true",
            help="Print some system information.",
        )
        self.add_option(
            "--transport",
            default="zeromq",
            choices=("zeromq", "tcp"),
            help=(
                "Select which transport to run the integration tests with, "
                "zeromq or tcp. Default: %default"
            ),
        )
        self.add_option(
            "--interactive",
            default=False,
            action="store_true",
            help="Do not run any tests. Simply start the daemons.",
        )
        self.output_options_group.add_option(
            "--no-colors",
            "--no-colours",
            default=False,
            action="store_true",
            help="Disable colour printing.",
        )

        self.test_selection_group.add_option(
            "-m",
            "--module",
            "--module-tests",
            dest="module",
            default=False,
            action="store_true",
            help="Run tests for modules",
        )
        self.test_selection_group.add_option(
            "-S",
            "--state",
            "--state-tests",
            dest="state",
            default=False,
            action="store_true",
            help="Run tests for states",
        )
        self.test_selection_group.add_option(
            "-C",
            "--cli",
            "--cli-tests",
            dest="cli",
            default=False,
            action="store_true",
            help="Run tests for cli",
        )
        self.test_selection_group.add_option(
            "-c",
            "--client",
            "--client-tests",
            dest="client",
            default=False,
            action="store_true",
            help="Run tests for client",
        )
        self.test_selection_group.add_option(
            "-d",
            "--doc",
            "--doc-tests",
            dest="doc",
            default=False,
            action="store_true",
            help="Run tests for documentation",
        )
        self.test_selection_group.add_option(
            "-I",
            "--ext-pillar",
            "--ext-pillar-tests",
            dest="ext_pillar",
            default=False,
            action="store_true",
            help="Run ext_pillar tests",
        )
        self.test_selection_group.add_option(
            "-G",
            "--grains",
            "--grains-tests",
            dest="grains",
            default=False,
            action="store_true",
            help="Run tests for grains",
        )
        self.test_selection_group.add_option(
            "-s",
            "--shell",
            "--shell-tests",
            dest="shell",
            default=False,
            action="store_true",
            help="Run shell tests",
        )
        self.test_selection_group.add_option(
            "-r",
            "--runners",
            "--runner-tests",
            dest="runners",
            default=False,
            action="store_true",
            help="Run salt/runners/*.py tests",
        )
        self.test_selection_group.add_option(
            "-R",
            "--renderers",
            "--renderer-tests",
            dest="renderers",
            default=False,
            action="store_true",
            help="Run salt/renderers/*.py tests",
        )
        self.test_selection_group.add_option(
            "--reactor",
            dest="reactor",
            default=False,
            action="store_true",
            help="Run salt/reactor/*.py tests",
        )
        self.test_selection_group.add_option(
            "--minion",
            "--minion-tests",
            dest="minion",
            default=False,
            action="store_true",
            help="Run tests for minion",
        )
        self.test_selection_group.add_option(
            "--returners",
            dest="returners",
            default=False,
            action="store_true",
            help="Run salt/returners/*.py tests",
        )
        self.test_selection_group.add_option(
            "--spm",
            dest="spm",
            default=False,
            action="store_true",
            help="Run spm integration tests",
        )
        self.test_selection_group.add_option(
            "--setup",
            dest="setup",
            default=False,
            action="store_true",
            help="Run setup integration tests",
        )
        self.test_selection_group.add_option(
            "-l",
            "--loader",
            "--loader-tests",
            dest="loader",
            default=False,
            action="store_true",
            help="Run loader tests",
        )
        self.test_selection_group.add_option(
            "-u",
            "--unit",
            "--unit-tests",
            dest="unit",
            default=False,
            action="store_true",
            help="Run unit tests",
        )
        self.test_selection_group.add_option(
            "-k",
            "--kitchen",
            "--kitchen-tests",
            dest="kitchen",
            default=False,
            action="store_true",
            help="Run kitchen tests",
        )
        self.test_selection_group.add_option(
            "--fileserver",
            "--fileserver-tests",
            dest="fileserver",
            default=False,
            action="store_true",
            help="Run Fileserver tests",
        )
        self.test_selection_group.add_option(
            "-w",
            "--wheel",
            "--wheel-tests",
            dest="wheel",
            action="store_true",
            default=False,
            help="Run wheel tests",
        )
        self.test_selection_group.add_option(
            "-o",
            "--outputter",
            "--outputter-tests",
            dest="outputter",
            action="store_true",
            default=False,
            help="Run outputter tests",
        )
        self.test_selection_group.add_option(
            "--cloud-provider",
            "--cloud-provider-tests",
            dest="cloud_provider",
            action="store_true",
            default=False,
            help=(
                "Run cloud provider tests. These tests create and delete "
                "instances on cloud providers. Must provide valid credentials "
                "in salt/tests/integration/files/conf/cloud.*.d to run tests."
            ),
        )
        self.test_selection_group.add_option(
            "--ssh",
            "--ssh-tests",
            dest="ssh",
            action="store_true",
            default=False,
            help="Run salt-ssh tests. These tests will spin up a temporary "
            "SSH server on your machine. In certain environments, this "
            "may be insecure! Default: False",
        )
        self.test_selection_group.add_option(
            "--ssh-int",
            dest="ssh-int",
            action="store_true",
            default=False,
            help="Run salt-ssh integration tests. Requires to be run with --ssh"
            "to spin up the SSH server on your machine.",
        )
        self.test_selection_group.add_option(
            "-A",
            "--api",
            "--api-tests",
            dest="api",
            action="store_true",
            default=False,
            help="Run salt-api tests",
        )
        self.test_selection_group.add_option(
            "--sdb",
            "--sdb-tests",
            dest="sdb",
            action="store_true",
            default=False,
            help="Run sdb tests",
        )
        self.test_selection_group.add_option(
            "-P",
            "--proxy",
            "--proxy-tests",
            dest="proxy",
            action="store_true",
            default=False,
            help="Run salt-proxy tests",
        )
        self.test_selection_group.add_option(
            "--external",
            "--external-api",
            "--external-api-tests",
            dest="external_api",
            action="store_true",
            default=False,
            help="Run venafi runner tests",
        )
        self.test_selection_group.add_option(
            "--daemons",
            "--daemon-tests",
            dest="daemons",
            action="store_true",
            default=False,
            help="Run salt/daemons/*.py tests",
        )
        self.test_selection_group.add_option(
            "--scheduler",
            dest="scheduler",
            action="store_true",
            default=False,
            help="Run scheduler integration tests",
        )
        self.test_selection_group.add_option(
            "--logging",
            dest="logging",
            action="store_true",
            default=False,
            help="Run logging integration tests",
        )
        self.test_selection_group.add_option(
            "--multimaster",
            dest="multimaster",
            action="store_true",
            default=False,
            help="Start multimaster daemons and run multimaster integration tests",
        )

    def validate_options(self):
        if self.options.cloud_provider or self.options.external_api:
            # Turn on expensive tests execution
            os.environ["EXPENSIVE_TESTS"] = "True"

        # This fails even with salt.utils.platform imported in the global
        # scope, unless we import it again here.
        import salt.utils.platform

        if salt.utils.platform.is_windows():
            import salt.utils.win_functions

            current_user = salt.utils.win_functions.get_current_user()
            if current_user == "SYSTEM":
                is_admin = True
            else:
                is_admin = salt.utils.win_functions.is_admin(current_user)
            if (
                self.options.coverage
                and any(
                    (self.options.name, not is_admin, not self.options.run_destructive)
                )
                and self._check_enabled_suites(include_unit=True)
            ):
                warnings.warn("Test suite not running with elevated priviledges")
        else:
            is_admin = os.geteuid() == 0

            if (
                self.options.coverage
                and any(
                    (self.options.name, not is_admin, not self.options.run_destructive)
                )
                and self._check_enabled_suites(include_unit=True)
            ):
                self.error(
                    "No sense in generating the tests coverage report when "
                    "not running the full test suite, including the "
                    "destructive tests, as 'root'. It would only produce "
                    "incorrect results."
                )

        # When no tests are specifically enumerated on the command line, setup
        # a default run: +unit -cloud_provider
        if not self.options.name and not self._check_enabled_suites(
            include_unit=True,
            include_cloud_provider=True,
            include_proxy=True,
            include_kitchen=True,
            include_multimaster=True,
        ):
            self._enable_suites(include_unit=True, include_multimaster=True)

        self.start_coverage(
            branch=True, source=[os.path.join(SALT_ROOT, "salt")],
        )

        # Print out which version of python this test suite is running on
        print(" * Python Version: {0}".format(" ".join(sys.version.split())))

        # Transplant configuration
        TestDaemon.transplant_configs(transport=self.options.transport)
        MultimasterTestDaemon.transplant_configs(transport=self.options.transport)

    def post_execution_cleanup(self):
        SaltCoverageTestingParser.post_execution_cleanup(self)
        if self.options.clean:
            TestDaemon.clean()

    def run_integration_suite(self, path="", display_name=""):
        """
        Run an integration test suite
        """
        full_path = os.path.join(TEST_DIR, path)
        return self.run_suite(
            full_path, display_name, suffix="test_*.py", failfast=self.options.failfast,
        )

    def start_daemons_only(self):
        if not salt.utils.platform.is_windows():
            self.set_filehandle_limits("integration")
        try:
            print_header(
                " * Setting up Salt daemons for interactive use",
                top=False,
                width=getattr(self.options, "output_columns", PNUM),
            )
        except TypeError:
            print_header(" * Setting up Salt daemons for interactive use", top=False)

        try:
            with TestDaemon(self):
                print_header(" * Salt daemons started")
                master_conf = TestDaemon.config("master")
                minion_conf = TestDaemon.config("minion")
                proxy_conf = TestDaemon.config("proxy")
                sub_minion_conf = TestDaemon.config("sub_minion")
                syndic_conf = TestDaemon.config("syndic")
                syndic_master_conf = TestDaemon.config("syndic_master")

                print_header(" * Syndic master configuration values (MoM)", top=False)
                print("interface: {0}".format(syndic_master_conf["interface"]))
                print("publish port: {0}".format(syndic_master_conf["publish_port"]))
                print("return port: {0}".format(syndic_master_conf["ret_port"]))
                print("\n")

                print_header(" * Syndic configuration values", top=True)
                print("interface: {0}".format(syndic_conf["interface"]))
                print("syndic master: {0}".format(syndic_conf["syndic_master"]))
                print(
                    "syndic master port: {0}".format(syndic_conf["syndic_master_port"])
                )
                print("\n")

                print_header(" * Master configuration values", top=True)
                print("interface: {0}".format(master_conf["interface"]))
                print("publish port: {0}".format(master_conf["publish_port"]))
                print("return port: {0}".format(master_conf["ret_port"]))
                print("\n")

                print_header(" * Minion configuration values", top=True)
                print("interface: {0}".format(minion_conf["interface"]))
                print("master: {0}".format(minion_conf["master"]))
                print("master port: {0}".format(minion_conf["master_port"]))
                if minion_conf["ipc_mode"] == "tcp":
                    print("tcp pub port: {0}".format(minion_conf["tcp_pub_port"]))
                    print("tcp pull port: {0}".format(minion_conf["tcp_pull_port"]))
                print("\n")

                print_header(" * Sub Minion configuration values", top=True)
                print("interface: {0}".format(sub_minion_conf["interface"]))
                print("master: {0}".format(sub_minion_conf["master"]))
                print("master port: {0}".format(sub_minion_conf["master_port"]))
                if sub_minion_conf["ipc_mode"] == "tcp":
                    print("tcp pub port: {0}".format(sub_minion_conf["tcp_pub_port"]))
                    print("tcp pull port: {0}".format(sub_minion_conf["tcp_pull_port"]))
                print("\n")

                print_header(" * Proxy Minion configuration values", top=True)
                print("interface: {0}".format(proxy_conf["interface"]))
                print("master: {0}".format(proxy_conf["master"]))
                print("master port: {0}".format(proxy_conf["master_port"]))
                if minion_conf["ipc_mode"] == "tcp":
                    print("tcp pub port: {0}".format(proxy_conf["tcp_pub_port"]))
                    print("tcp pull port: {0}".format(proxy_conf["tcp_pull_port"]))
                print("\n")

                print_header(
                    " Your client configuration is at {0}".format(
                        TestDaemon.config_location()
                    )
                )
                print(
                    "To access the minion: salt -c {0} minion test.ping".format(
                        TestDaemon.config_location()
                    )
                )

                while True:
                    time.sleep(1)
        except TestDaemonStartFailed:
            self.exit(status=2)

    def start_multimaster_daemons_only(self):
        if not salt.utils.platform.is_windows():
            self.set_filehandle_limits("integration")
        try:
            print_header(
                " * Setting up Salt daemons for interactive use",
                top=False,
                width=getattr(self.options, "output_columns", PNUM),
            )
        except TypeError:
            print_header(" * Setting up Salt daemons for interactive use", top=False)

        try:
            with MultimasterTestDaemon(self):
                print_header(" * Salt daemons started")
                master_conf = MultimasterTestDaemon.config("mm_master")
                sub_master_conf = MultimasterTestDaemon.config("mm_sub_master")
                minion_conf = MultimasterTestDaemon.config("mm_minion")
                sub_minion_conf = MultimasterTestDaemon.config("mm_sub_minion")

                print_header(" * Master configuration values", top=True)
                print("interface: {0}".format(master_conf["interface"]))
                print("publish port: {0}".format(master_conf["publish_port"]))
                print("return port: {0}".format(master_conf["ret_port"]))
                print("\n")

                print_header(" * Second master configuration values", top=True)
                print("interface: {0}".format(sub_master_conf["interface"]))
                print("publish port: {0}".format(sub_master_conf["publish_port"]))
                print("return port: {0}".format(sub_master_conf["ret_port"]))
                print("\n")

                print_header(" * Minion configuration values", top=True)
                print("interface: {0}".format(minion_conf["interface"]))
                print("masters: {0}".format(", ".join(minion_conf["master"])))
                if minion_conf["ipc_mode"] == "tcp":
                    print("tcp pub port: {0}".format(minion_conf["tcp_pub_port"]))
                    print("tcp pull port: {0}".format(minion_conf["tcp_pull_port"]))
                print("\n")

                print_header(" * Sub Minion configuration values", top=True)
                print("interface: {0}".format(sub_minion_conf["interface"]))
                print("masters: {0}".format(", ".join(sub_minion_conf["master"])))
                if sub_minion_conf["ipc_mode"] == "tcp":
                    print("tcp pub port: {0}".format(sub_minion_conf["tcp_pub_port"]))
                    print("tcp pull port: {0}".format(sub_minion_conf["tcp_pull_port"]))
                print("\n")

                print_header(
                    " Your client configurations are at {0}".format(
                        ", ".join(MultimasterTestDaemon.config_location())
                    )
                )
                print("To access minions from different masters use:")
                for location in MultimasterTestDaemon.config_location():
                    print("    salt -c {0} minion test.ping".format(location))

                while True:
                    time.sleep(1)
        except TestDaemonStartFailed:
            self.exit(status=2)

    def set_filehandle_limits(self, limits="integration"):
        """
        Set soft and hard limits on open file handles at required thresholds
        for integration tests or unit tests
        """
        # Get current limits
        if salt.utils.platform.is_windows():
            import win32file

            prev_hard = win32file._getmaxstdio()
            prev_soft = 512
        else:
            prev_soft, prev_hard = resource.getrlimit(resource.RLIMIT_NOFILE)

        # Get required limits
        min_soft = MAX_OPEN_FILES[limits]["soft_limit"]
        min_hard = MAX_OPEN_FILES[limits]["hard_limit"]

        # Check minimum required limits
        set_limits = False
        if prev_soft < min_soft:
            soft = min_soft
            set_limits = True
        else:
            soft = prev_soft

        if prev_hard < min_hard:
            hard = min_hard
            set_limits = True
        else:
            hard = prev_hard

        # Increase limits
        if set_limits:
            print(
                " * Max open files settings is too low (soft: {0}, hard: {1}) "
                "for running the tests".format(prev_soft, prev_hard)
            )
            print(
                " * Trying to raise the limits to soft: "
                "{0}, hard: {1}".format(soft, hard)
            )
            try:
                if salt.utils.platform.is_windows():
                    hard = 2048 if hard > 2048 else hard
                    win32file._setmaxstdio(hard)
                else:
                    resource.setrlimit(resource.RLIMIT_NOFILE, (soft, hard))
            except Exception as err:  # pylint: disable=broad-except
                print(
                    "ERROR: Failed to raise the max open files settings -> "
                    "{0}".format(err)
                )
                print("Please issue the following command on your console:")
                print("  ulimit -n {0}".format(soft))
                self.exit()
            finally:
                print("~" * getattr(self.options, "output_columns", PNUM))

    def run_integration_tests(self):
        """
        Execute the integration tests suite
        """
        named_tests = []
        named_unit_test = []

        if self.options.name:
            for test in self.options.name:
                if test.startswith(
                    (
                        "tests.unit.",
                        "unit.",
                        "test.kitchen.",
                        "kitchen.",
                        "test.multimaster.",
                        "multimaster.",
                    )
                ):
                    named_unit_test.append(test)
                    continue
                named_tests.append(test)

        if (
            (
                self.options.unit
                or self.options.kitchen
                or self.options.multimaster
                or named_unit_test
            )
            and not named_tests
            and (
                self.options.from_filenames
                or not self._check_enabled_suites(include_cloud_provider=True)
            )
        ):
            # We're either not running any integration test suites, or we're
            # only running unit tests by passing --unit or by passing only
            # `unit.<whatever>` to --name.  We don't need the tests daemon
            # running
            return [True]

        if not salt.utils.platform.is_windows():
            self.set_filehandle_limits("integration")

        try:
            print_header(
                " * Setting up Salt daemons to execute tests",
                top=False,
                width=getattr(self.options, "output_columns", PNUM),
            )
        except TypeError:
            print_header(" * Setting up Salt daemons to execute tests", top=False)

        status = []
        # Return an empty status if no tests have been enabled
        if (
            not self._check_enabled_suites(
                include_cloud_provider=True, include_proxy=True
            )
            and not self.options.name
        ):
            return status

        try:
            with TestDaemon(self):
                if self.options.name:
                    for name in self.options.name:
                        name = name.strip()
                        if not name:
                            continue
                        if os.path.isfile(name):
                            if not name.endswith(".py"):
                                continue
                            if name.startswith(
                                (
                                    os.path.join("tests", "unit"),
                                    os.path.join("tests", "multimaster"),
                                )
                            ):
                                continue
                            results = self.run_suite(
                                os.path.dirname(name),
                                name,
                                suffix=os.path.basename(name),
                                failfast=self.options.failfast,
                                load_from_name=False,
                            )
                            status.append(results)
                            continue
                        if name.startswith(
                            (
                                "tests.unit.",
                                "unit.",
                                "tests.multimaster.",
                                "multimaster.",
                            )
                        ):
                            continue
                        results = self.run_suite(
                            "",
                            name,
                            suffix="test_*.py",
                            load_from_name=True,
                            failfast=self.options.failfast,
                        )
                        status.append(results)
                    return status
                for suite in TEST_SUITES:
                    if (
                        suite != "unit"
                        and suite != "multimaster"
                        and getattr(self.options, suite)
                    ):
                        status.append(self.run_integration_suite(**TEST_SUITES[suite]))
            return status
        except TestDaemonStartFailed:
            self.exit(status=2)

    def run_multimaster_tests(self):
        """
        Execute the multimaster tests suite
        """
        named_tests = []
        named_unit_test = []

        if self.options.name:
            for test in self.options.name:
                if test.startswith(("tests.multimaster.", "multimaster.")):
                    named_tests.append(test)

        # TODO: check 'from_filenames'
        if not self.options.multimaster and not named_tests:
            # We're not running any multimaster test suites.
            return [True]

        if not salt.utils.platform.is_windows():
            self.set_filehandle_limits("integration")

        try:
            print_header(
                " * Setting up multimaster Salt daemons to execute tests",
                top=False,
                width=getattr(self.options, "output_columns", PNUM),
            )
        except TypeError:
            print_header(
                " * Setting up multimaster Salt daemons to execute tests", top=False
            )

        status = []

        try:
            with MultimasterTestDaemon(self):
                if self.options.name:
                    for name in self.options.name:
                        name = name.strip()
                        if not name:
                            continue
                        if os.path.isfile(name):
                            if not name.endswith(".py"):
                                continue
                            if not name.startswith(
                                os.path.join("tests", "multimaster")
                            ):
                                continue
                            results = self.run_suite(
                                os.path.dirname(name),
                                name,
                                suffix=os.path.basename(name),
                                load_from_name=False,
                            )
                            status.append(results)
                            continue
                        if not name.startswith(("tests.multimaster.", "multimaster.")):
                            continue
                        results = self.run_suite(
                            "", name, suffix="test_*.py", load_from_name=True
                        )
                        status.append(results)
                    return status
                status.append(self.run_integration_suite(**TEST_SUITES["multimaster"]))
            return status
        except TestDaemonStartFailed:
            self.exit(status=2)

    def run_unit_tests(self):
        """
        Execute the unit tests
        """
        named_unit_test = []
        if self.options.name:
            for test in self.options.name:
                if not test.startswith(("tests.unit.", "unit.")):
                    continue
                named_unit_test.append(test)

        if not named_unit_test and (
            self.options.from_filenames or not self.options.unit
        ):
            # We are not explicitly running the unit tests and none of the
            # names passed to --name (or derived via --from-filenames) is a
            # unit test.
            return [True]

        status = []
        if self.options.unit:
            # MacOS needs more open filehandles for running unit test suite
            self.set_filehandle_limits("unit")

            results = self.run_suite(
                os.path.join(TEST_DIR, "unit"),
                "Unit",
                suffix="test_*.py",
                failfast=self.options.failfast,
            )
            status.append(results)
            # We executed ALL unittests, we can skip running unittests by name
            # below
            return status

        for name in named_unit_test:
            results = self.run_suite(
                os.path.join(TEST_DIR, "unit"),
                name,
                suffix="test_*.py",
                load_from_name=True,
                failfast=self.options.failfast,
            )
            status.append(results)
        return status

    def run_kitchen_tests(self):
        """
        Execute the kitchen tests
        """
        named_kitchen_test = []
        if self.options.name:
            for test in self.options.name:
                if not test.startswith(("tests.kitchen.", "kitchen.")):
                    continue
                named_kitchen_test.append(test)

        if not self.options.kitchen and not named_kitchen_test:
            # We are not explicitly running the unit tests and none of the
            # names passed to --name is a unit test.
            return [True]

        status = []
        if self.options.kitchen:
            results = self.run_suite(
                os.path.join(TEST_DIR, "kitchen"), "Kitchen", suffix="test_*.py"
            )
            status.append(results)
            # We executed ALL unittests, we can skip running unittests by name
            # below
            return status

        for name in named_kitchen_test:
            results = self.run_suite(
                os.path.join(TEST_DIR, "kitchen"),
                name,
                suffix="test_*.py",
                load_from_name=True,
            )
            status.append(results)
        return status


def main(**kwargs):
    """
    Parse command line options for running specific tests
    """
    try:
        parser = SaltTestsuiteParser(
            TEST_DIR,
            xml_output_dir=XML_OUTPUT_DIR,
            tests_logfile=os.path.join(SYS_TMP_DIR, "salt-runtests.log"),
        )
        parser.parse_args()

        # Override parser options (helpful when importing runtests.py and
        # running from within a REPL). Using kwargs.items() to avoid importing
        # six, as this feature will rarely be used.
        for key, val in kwargs.items():
            setattr(parser.options, key, val)

        overall_status = []
        if parser.options.interactive:
            if parser.options.multimaster:
                parser.start_multimaster_daemons_only()
            else:
                parser.start_daemons_only()
        status = parser.run_integration_tests()
        overall_status.extend(status)
        status = parser.run_multimaster_tests()
        overall_status.extend(status)
        status = parser.run_unit_tests()
        overall_status.extend(status)
        status = parser.run_kitchen_tests()
        overall_status.extend(status)
        false_count = overall_status.count(False)

        if false_count > 0:
            parser.finalize(1)
        parser.finalize(0)
    except KeyboardInterrupt:
        print("\nCaught keyboard interrupt. Exiting.\n")
        exit(0)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
"""
Tests for salt.states.zfs

:codeauthor:    Jorge Schrauwen <sjorge@blackdot.be>
:maintainer:    Jorge Schrauwen <sjorge@blackdot.be>
:maturity:      new
:depends:       salt.utils.zfs, salt.modules.zfs
:platform:      illumos,freebsd,linux
"""
from __future__ import absolute_import, print_function, unicode_literals

import salt.loader
import salt.states.zfs as zfs
import salt.utils.zfs
from salt.utils.odict import OrderedDict
from tests.support.helpers import slowTest
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase
from tests.support.zfs import ZFSMockData


class ZfsTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.states.zfs
    """

    def setup_loader_modules(self):
        self.opts = opts = salt.config.DEFAULT_MINION_OPTS.copy()
        self.utils_patch = ZFSMockData().get_patched_utils()
        for key in ("opts", "utils_patch"):
            self.addCleanup(delattr, self, key)
        utils = salt.loader.utils(opts, whitelist=["zfs"])
        zfs_obj = {
            zfs: {
                "__opts__": opts,
                "__grains__": {"kernel": "SunOS"},
                "__utils__": utils,
            }
        }

        return zfs_obj

    def test_filesystem_absent_nofs(self):
        """
        Test if filesystem is absent (non existing filesystem)
        """
        ret = {
            "name": "myzpool/filesystem",
            "result": True,
            "comment": "filesystem myzpool/filesystem is absent",
            "changes": {},
        }

        mock_exists = MagicMock(return_value=False)
        with patch.dict(zfs.__salt__, {"zfs.exists": mock_exists}), patch.dict(
            zfs.__utils__, self.utils_patch
        ):
            self.assertEqual(ret, zfs.filesystem_absent("myzpool/filesystem"))

    def test_filesystem_absent_removed(self):
        """
        Test if filesystem is absent
        """
        ret = {
            "name": "myzpool/filesystem",
            "result": True,
            "comment": "filesystem myzpool/filesystem was destroyed",
            "changes": {"myzpool/filesystem": "destroyed"},
        }

        mock_exists = MagicMock(return_value=True)
        mock_destroy = MagicMock(return_value=OrderedDict([("destroyed", True)]))
        with patch.dict(zfs.__salt__, {"zfs.exists": mock_exists}), patch.dict(
            zfs.__salt__, {"zfs.destroy": mock_destroy}
        ), patch.dict(zfs.__utils__, self.utils_patch):
            self.assertEqual(ret, zfs.filesystem_absent("myzpool/filesystem"))

    def test_filesystem_absent_fail(self):
        """
        Test if filesystem is absent (with snapshots)
        """
        ret = {
            "name": "myzpool/filesystem",
            "result": False,
            "comment": "\n".join(
                [
                    "cannot destroy 'myzpool/filesystem': filesystem has children",
                    "use 'recursive=True' to destroy the following datasets:",
                    "myzpool/filesystem@snap",
                ]
            ),
            "changes": {},
        }

        mock_exists = MagicMock(return_value=True)
        mock_destroy = MagicMock(
            return_value=OrderedDict(
                [
                    ("destroyed", False),
                    (
                        "error",
                        "\n".join(
                            [
                                "cannot destroy 'myzpool/filesystem': filesystem has children",
                                "use 'recursive=True' to destroy the following datasets:",
                                "myzpool/filesystem@snap",
                            ]
                        ),
                    ),
                ]
            )
        )
        with patch.dict(zfs.__salt__, {"zfs.exists": mock_exists}), patch.dict(
            zfs.__salt__, {"zfs.destroy": mock_destroy}
        ), patch.dict(zfs.__utils__, self.utils_patch):
            self.assertEqual(ret, zfs.filesystem_absent("myzpool/filesystem"))

    def test_volume_absent_novol(self):
        """
        Test if volume is absent (non existing volume)
        """
        ret = {
            "name": "myzpool/volume",
            "result": True,
            "comment": "volume myzpool/volume is absent",
            "changes": {},
        }

        mock_exists = MagicMock(return_value=False)
        with patch.dict(zfs.__salt__, {"zfs.exists": mock_exists}), patch.dict(
            zfs.__utils__, self.utils_patch
        ):
            self.assertEqual(ret, zfs.volume_absent("myzpool/volume"))

    def test_volume_absent_removed(self):
        """
        Test if volume is absent
        """
        ret = {
            "name": "myzpool/volume",
            "result": True,
            "comment": "volume myzpool/volume was destroyed",
            "changes": {"myzpool/volume": "destroyed"},
        }

        mock_exists = MagicMock(return_value=True)
        mock_destroy = MagicMock(return_value=OrderedDict([("destroyed", True)]))
        with patch.dict(zfs.__salt__, {"zfs.exists": mock_exists}), patch.dict(
            zfs.__salt__, {"zfs.destroy": mock_destroy}
        ), patch.dict(zfs.__utils__, self.utils_patch):
            self.assertEqual(ret, zfs.volume_absent("myzpool/volume"))

    def test_volume_absent_fail(self):
        """
        Test if volume is absent (with snapshots)
        """
        ret = {
            "name": "myzpool/volume",
            "result": False,
            "comment": "\n".join(
                [
                    "cannot destroy 'myzpool/volume': volume has children",
                    "use 'recursive=True' to destroy the following datasets:",
                    "myzpool/volume@snap",
                ]
            ),
            "changes": {},
        }

        mock_exists = MagicMock(return_value=True)
        mock_destroy = MagicMock(
            return_value=OrderedDict(
                [
                    ("destroyed", False),
                    (
                        "error",
                        "\n".join(
                            [
                                "cannot destroy 'myzpool/volume': volume has children",
                                "use 'recursive=True' to destroy the following datasets:",
                                "myzpool/volume@snap",
                            ]
                        ),
                    ),
                ]
            )
        )
        with patch.dict(zfs.__salt__, {"zfs.exists": mock_exists}), patch.dict(
            zfs.__salt__, {"zfs.destroy": mock_destroy}
        ), patch.dict(zfs.__utils__, self.utils_patch):
            self.assertEqual(ret, zfs.volume_absent("myzpool/volume"))

    def test_snapshot_absent_nosnap(self):
        """
        Test if snapshot is absent (non existing snapshot)
        """
        ret = {
            "name": "myzpool/filesystem@snap",
            "result": True,
            "comment": "snapshot myzpool/filesystem@snap is absent",
            "changes": {},
        }

        mock_exists = MagicMock(return_value=False)
        with patch.dict(zfs.__salt__, {"zfs.exists": mock_exists}), patch.dict(
            zfs.__utils__, self.utils_patch
        ):
            self.assertEqual(ret, zfs.snapshot_absent("myzpool/filesystem@snap"))

    def test_snapshot_absent_removed(self):
        """
        Test if snapshot is absent
        """
        ret = {
            "name": "myzpool/filesystem@snap",
            "result": True,
            "comment": "snapshot myzpool/filesystem@snap was destroyed",
            "changes": {"myzpool/filesystem@snap": "destroyed"},
        }

        mock_exists = MagicMock(return_value=True)
        mock_destroy = MagicMock(return_value=OrderedDict([("destroyed", True)]))
        with patch.dict(zfs.__salt__, {"zfs.exists": mock_exists}), patch.dict(
            zfs.__salt__, {"zfs.destroy": mock_destroy}
        ), patch.dict(zfs.__utils__, self.utils_patch):
            self.assertEqual(ret, zfs.snapshot_absent("myzpool/filesystem@snap"))

    def test_snapshot_absent_fail(self):
        """
        Test if snapshot is absent (with snapshots)
        """
        ret = {
            "name": "myzpool/filesystem@snap",
            "result": False,
            "comment": "cannot destroy snapshot myzpool/filesystem@snap: dataset is busy",
            "changes": {},
        }

        mock_exists = MagicMock(return_value=True)
        mock_destroy = MagicMock(
            return_value=OrderedDict(
                [
                    ("destroyed", False),
                    (
                        "error",
                        "cannot destroy snapshot myzpool/filesystem@snap: dataset is busy",
                    ),
                ]
            )
        )
        with patch.dict(zfs.__salt__, {"zfs.exists": mock_exists}), patch.dict(
            zfs.__salt__, {"zfs.destroy": mock_destroy}
        ), patch.dict(zfs.__utils__, self.utils_patch):
            self.assertEqual(ret, zfs.snapshot_absent("myzpool/filesystem@snap"))

    @slowTest
    def test_bookmark_absent_nobook(self):
        """
        Test if bookmark is absent (non existing bookmark)
        """
        ret = {
            "name": "myzpool/filesystem#book",
            "result": True,
            "comment": "bookmark myzpool/filesystem#book is absent",
            "changes": {},
        }

        mock_exists = MagicMock(return_value=False)
        with patch.dict(zfs.__salt__, {"zfs.exists": mock_exists}), patch.dict(
            zfs.__utils__, self.utils_patch
        ):
            self.assertEqual(ret, zfs.bookmark_absent("myzpool/filesystem#book"))

    def test_bookmark_absent_removed(self):
        """
        Test if bookmark is absent
        """
        ret = {
            "name": "myzpool/filesystem#book",
            "result": True,
            "comment": "bookmark myzpool/filesystem#book was destroyed",
            "changes": {"myzpool/filesystem#book": "destroyed"},
        }

        mock_exists = MagicMock(return_value=True)
        mock_destroy = MagicMock(return_value=OrderedDict([("destroyed", True)]))
        with patch.dict(zfs.__salt__, {"zfs.exists": mock_exists}), patch.dict(
            zfs.__salt__, {"zfs.destroy": mock_destroy}
        ), patch.dict(zfs.__utils__, self.utils_patch):
            self.assertEqual(ret, zfs.bookmark_absent("myzpool/filesystem#book"))

    def test_hold_absent_nohold(self):
        """
        Test if hold is absent (non existing hold)
        """
        ret = {
            "name": "myhold",
            "result": True,
            "comment": "hold myhold is absent",
            "changes": {},
        }

        mock_holds = MagicMock(return_value=OrderedDict([]))
        with patch.dict(zfs.__salt__, {"zfs.holds": mock_holds}), patch.dict(
            zfs.__utils__, self.utils_patch
        ):
            self.assertEqual(ret, zfs.hold_absent("myhold", "myzpool/filesystem@snap"))

    @slowTest
    def test_hold_absent_removed(self):
        """
        Test if hold is absent
        """
        ret = {
            "name": "myhold",
            "result": True,
            "comment": "hold myhold released",
            "changes": OrderedDict(
                [("myzpool/filesystem@snap", OrderedDict([("myhold", "released")]))]
            ),
        }

        mock_holds = MagicMock(
            return_value=OrderedDict([("myhold", "Thu Feb 15 16:24 2018")])
        )
        mock_release = MagicMock(return_value=OrderedDict([("released", True)]))
        with patch.dict(zfs.__salt__, {"zfs.holds": mock_holds}), patch.dict(
            zfs.__salt__, {"zfs.release": mock_release}
        ), patch.dict(zfs.__utils__, self.utils_patch):
            self.assertEqual(ret, zfs.hold_absent("myhold", "myzpool/filesystem@snap"))

    def test_hold_absent_fail(self):
        """
        Test if hold is absent (non existing snapshot)
        """
        ret = {
            "name": "myhold",
            "result": False,
            "comment": "cannot open 'myzpool/filesystem@snap': dataset does not exist",
            "changes": {},
        }

        mock_holds = MagicMock(
            return_value=OrderedDict(
                [
                    (
                        "error",
                        "cannot open 'myzpool/filesystem@snap': dataset does not exist",
                    ),
                ]
            )
        )
        with patch.dict(zfs.__salt__, {"zfs.holds": mock_holds}), patch.dict(
            zfs.__utils__, self.utils_patch
        ):
            self.assertEqual(ret, zfs.hold_absent("myhold", "myzpool/filesystem@snap"))

    def test_hold_present(self):
        """
        Test if hold is present (hold already present)
        """
        ret = {
            "name": "myhold",
            "result": True,
            "comment": "hold myhold is present for myzpool/filesystem@snap",
            "changes": {},
        }

        mock_holds = MagicMock(
            return_value=OrderedDict([("myhold", "Thu Feb 15 16:24 2018")])
        )
        with patch.dict(zfs.__salt__, {"zfs.holds": mock_holds}), patch.dict(
            zfs.__utils__, self.utils_patch
        ):
            self.assertEqual(ret, zfs.hold_present("myhold", "myzpool/filesystem@snap"))

    @slowTest
    def test_hold_present_new(self):
        """
        Test if hold is present (new)
        """
        ret = {
            "name": "myhold",
            "result": True,
            "comment": "hold myhold added to myzpool/filesystem@snap",
            "changes": {"myzpool/filesystem@snap": {"myhold": "held"}},
        }

        mock_holds = MagicMock(return_value=OrderedDict([]))
        mock_hold = MagicMock(return_value=OrderedDict([("held", True)]))
        with patch.dict(zfs.__salt__, {"zfs.holds": mock_holds}), patch.dict(
            zfs.__salt__, {"zfs.hold": mock_hold}
        ), patch.dict(zfs.__utils__, self.utils_patch):
            self.assertEqual(ret, zfs.hold_present("myhold", "myzpool/filesystem@snap"))

    @slowTest
    def test_hold_present_fail(self):
        """
        Test if hold is present (using non existing snapshot)
        """
        ret = {
            "name": "myhold",
            "result": False,
            "comment": "cannot hold snapshot 'zsalt/filesystem@snap': dataset does not exist",
            "changes": {},
        }

        mock_holds = MagicMock(return_value=OrderedDict([]))
        mock_hold = MagicMock(
            return_value=OrderedDict(
                [
                    ("held", False),
                    (
                        "error",
                        "cannot hold snapshot 'zsalt/filesystem@snap': dataset does not exist",
                    ),
                ]
            )
        )
        with patch.dict(zfs.__salt__, {"zfs.holds": mock_holds}), patch.dict(
            zfs.__salt__, {"zfs.hold": mock_hold}
        ), patch.dict(zfs.__utils__, self.utils_patch):
            self.assertEqual(ret, zfs.hold_present("myhold", "myzpool/filesystem@snap"))

    @slowTest
    def test_filesystem_present(self):
        """
        Test if filesystem is present (existing filesystem)
        """
        ret = {
            "name": "myzpool/filesystem",
            "result": True,
            "comment": "filesystem myzpool/filesystem is uptodate",
            "changes": {},
        }

        mock_exists = MagicMock(return_value=True)
        mock_get = MagicMock(
            return_value=OrderedDict(
                [
                    (
                        "myzpool/filesystem",
                        OrderedDict(
                            [
                                ("type", OrderedDict([("value", "filesystem")])),
                                ("compression", OrderedDict([("value", False)])),
                            ]
                        ),
                    ),
                ]
            )
        )
        with patch.dict(zfs.__salt__, {"zfs.exists": mock_exists}), patch.dict(
            zfs.__salt__, {"zfs.get": mock_get}
        ), patch.dict(zfs.__utils__, self.utils_patch):
            self.assertEqual(ret, zfs.filesystem_present("myzpool/filesystem"))

    def test_filesystem_present_new(self):
        """
        Test if filesystem is present (non existing filesystem)
        """
        ret = {
            "name": "myzpool/filesystem",
            "result": True,
            "comment": "filesystem myzpool/filesystem was created",
            "changes": {"myzpool/filesystem": "created"},
        }

        mock_exists = MagicMock(return_value=False)
        mock_create = MagicMock(return_value=OrderedDict([("created", True)]))
        with patch.dict(zfs.__salt__, {"zfs.exists": mock_exists}), patch.dict(
            zfs.__salt__, {"zfs.create": mock_create}
        ), patch.dict(zfs.__utils__, self.utils_patch):
            self.assertEqual(ret, zfs.filesystem_present("myzpool/filesystem"))

    @slowTest
    def test_filesystem_present_update(self):
        """
        Test if filesystem is present (non existing filesystem)
        """
        ret = {
            "name": "myzpool/filesystem",
            "result": True,
            "comment": "filesystem myzpool/filesystem was updated",
            "changes": {"myzpool/filesystem": {"compression": "lz4"}},
        }

        mock_exists = MagicMock(return_value=True)
        mock_set = MagicMock(return_value=OrderedDict([("set", True)]))
        mock_get = MagicMock(
            return_value=OrderedDict(
                [
                    (
                        "myzpool/filesystem",
                        OrderedDict(
                            [
                                ("type", OrderedDict([("value", "filesystem")])),
                                ("compression", OrderedDict([("value", False)])),
                            ]
                        ),
                    ),
                ]
            )
        )
        with patch.dict(zfs.__salt__, {"zfs.exists": mock_exists}), patch.dict(
            zfs.__salt__, {"zfs.get": mock_get}
        ), patch.dict(zfs.__salt__, {"zfs.set": mock_set}), patch.dict(
            zfs.__utils__, self.utils_patch
        ):
            self.assertEqual(
                ret,
                zfs.filesystem_present(
                    name="myzpool/filesystem", properties={"compression": "lz4"},
                ),
            )

    def test_filesystem_present_fail(self):
        """
        Test if filesystem is present (non existing pool)
        """
        ret = {
            "name": "myzpool/filesystem",
            "result": False,
            "comment": "cannot create 'myzpool/filesystem': no such pool 'myzpool'",
            "changes": {},
        }

        mock_exists = MagicMock(return_value=False)
        mock_create = MagicMock(
            return_value=OrderedDict(
                [
                    ("created", False),
                    (
                        "error",
                        "cannot create 'myzpool/filesystem': no such pool 'myzpool'",
                    ),
                ]
            )
        )
        with patch.dict(zfs.__salt__, {"zfs.exists": mock_exists}), patch.dict(
            zfs.__salt__, {"zfs.create": mock_create}
        ), patch.dict(zfs.__utils__, self.utils_patch):
            self.assertEqual(ret, zfs.filesystem_present("myzpool/filesystem"))

    @slowTest
    def test_volume_present(self):
        """
        Test if volume is present (existing volume)
        """
        ret = {
            "name": "myzpool/volume",
            "result": True,
            "comment": "volume myzpool/volume is uptodate",
            "changes": {},
        }

        mock_exists = MagicMock(return_value=True)
        mock_get = MagicMock(
            return_value=OrderedDict(
                [
                    (
                        "myzpool/volume",
                        OrderedDict(
                            [
                                ("type", OrderedDict([("value", "volume")])),
                                ("compression", OrderedDict([("value", False)])),
                            ]
                        ),
                    ),
                ]
            )
        )
        with patch.dict(zfs.__salt__, {"zfs.exists": mock_exists}), patch.dict(
            zfs.__salt__, {"zfs.get": mock_get}
        ), patch.dict(zfs.__utils__, self.utils_patch):
            self.assertEqual(
                ret, zfs.volume_present("myzpool/volume", volume_size="1G")
            )

    def test_volume_present_new(self):
        """
        Test if volume is present (non existing volume)
        """
        ret = {
            "name": "myzpool/volume",
            "result": True,
            "comment": "volume myzpool/volume was created",
            "changes": {"myzpool/volume": "created"},
        }

        mock_exists = MagicMock(return_value=False)
        mock_create = MagicMock(return_value=OrderedDict([("created", True)]))
        with patch.dict(zfs.__salt__, {"zfs.exists": mock_exists}), patch.dict(
            zfs.__salt__, {"zfs.create": mock_create}
        ), patch.dict(zfs.__utils__, self.utils_patch):
            self.assertEqual(
                ret, zfs.volume_present("myzpool/volume", volume_size="1G")
            )

    def test_volume_present_update(self):
        """
        Test if volume is present (non existing volume)
        """
        ret = {
            "name": "myzpool/volume",
            "result": True,
            "comment": "volume myzpool/volume was updated",
            "changes": {"myzpool/volume": {"compression": "lz4"}},
        }

        mock_exists = MagicMock(return_value=True)
        mock_set = MagicMock(return_value=OrderedDict([("set", True)]))
        mock_get = MagicMock(
            return_value=OrderedDict(
                [
                    (
                        "myzpool/volume",
                        OrderedDict(
                            [
                                ("type", OrderedDict([("value", "volume")])),
                                ("compression", OrderedDict([("value", False)])),
                            ]
                        ),
                    ),
                ]
            )
        )
        with patch.dict(zfs.__salt__, {"zfs.exists": mock_exists}), patch.dict(
            zfs.__salt__, {"zfs.get": mock_get}
        ), patch.dict(zfs.__salt__, {"zfs.set": mock_set}), patch.dict(
            zfs.__utils__, self.utils_patch
        ):
            self.assertEqual(
                ret,
                zfs.volume_present(
                    name="myzpool/volume",
                    volume_size="1G",
                    properties={"compression": "lz4"},
                ),
            )

    def test_volume_present_fail(self):
        """
        Test if volume is present (non existing pool)
        """
        ret = {
            "name": "myzpool/volume",
            "result": False,
            "comment": "cannot create 'myzpool/volume': no such pool 'myzpool'",
            "changes": {},
        }

        mock_exists = MagicMock(return_value=False)
        mock_create = MagicMock(
            return_value=OrderedDict(
                [
                    ("created", False),
                    ("error", "cannot create 'myzpool/volume': no such pool 'myzpool'"),
                ]
            )
        )
        with patch.dict(zfs.__salt__, {"zfs.exists": mock_exists}), patch.dict(
            zfs.__salt__, {"zfs.create": mock_create}
        ), patch.dict(zfs.__utils__, self.utils_patch):
            self.assertEqual(
                ret, zfs.volume_present("myzpool/volume", volume_size="1G")
            )

    def test_bookmark_present(self):
        """
        Test if bookmark is present (bookmark already present)
        """
        ret = {
            "name": "myzpool/filesystem#mybookmark",
            "result": True,
            "comment": "bookmark is present",
            "changes": {},
        }

        mock_exists = MagicMock(return_value=True)
        with patch.dict(zfs.__salt__, {"zfs.exists": mock_exists}), patch.dict(
            zfs.__utils__, self.utils_patch
        ):
            self.assertEqual(
                ret, zfs.bookmark_present("mybookmark", "myzpool/filesystem@snap")
            )

    def test_bookmark_present_new(self):
        """
        Test if bookmark is present (new)
        """
        ret = {
            "name": "myzpool/filesystem#mybookmark",
            "result": True,
            "comment": "myzpool/filesystem@snap bookmarked as myzpool/filesystem#mybookmark",
            "changes": {"myzpool/filesystem#mybookmark": "myzpool/filesystem@snap"},
        }

        mock_exists = MagicMock(return_value=False)
        mock_bookmark = MagicMock(return_value=OrderedDict([("bookmarked", True)]))
        with patch.dict(zfs.__salt__, {"zfs.exists": mock_exists}), patch.dict(
            zfs.__salt__, {"zfs.bookmark": mock_bookmark}
        ), patch.dict(zfs.__utils__, self.utils_patch):
            self.assertEqual(
                ret, zfs.bookmark_present("mybookmark", "myzpool/filesystem@snap")
            )

    def test_bookmark_present_fail(self):
        """
        Test if bookmark is present (using non existing snapshot)
        """
        ret = {
            "name": "myzpool/filesystem#mybookmark",
            "result": False,
            "comment": "cannot bookmark snapshot 'zsalt/filesystem@snap': dataset does not exist",
            "changes": {},
        }

        mock_exists = MagicMock(return_value=False)
        mock_bookmark = MagicMock(
            return_value=OrderedDict(
                [
                    ("bookmarked", False),
                    (
                        "error",
                        "cannot bookmark snapshot 'zsalt/filesystem@snap': dataset does not exist",
                    ),
                ]
            )
        )
        with patch.dict(zfs.__salt__, {"zfs.exists": mock_exists}), patch.dict(
            zfs.__salt__, {"zfs.bookmark": mock_bookmark}
        ), patch.dict(zfs.__utils__, self.utils_patch):
            self.assertEqual(
                ret, zfs.bookmark_present("mybookmark", "myzpool/filesystem@snap")
            )

    @slowTest
    def test_snapshot_present(self):
        """
        Test if snapshot is present (snapshot already present)
        """
        ret = {
            "name": "myzpool/filesystem@snap",
            "result": True,
            "comment": "snapshot is present",
            "changes": {},
        }

        mock_exists = MagicMock(return_value=True)
        with patch.dict(zfs.__salt__, {"zfs.exists": mock_exists}), patch.dict(
            zfs.__utils__, self.utils_patch
        ):
            self.assertEqual(ret, zfs.snapshot_present("myzpool/filesystem@snap"))

    @slowTest
    def test_snapshot_present_new(self):
        """
        Test if snapshot is present (new)
        """
        ret = {
            "name": "myzpool/filesystem@snap",
            "result": True,
            "comment": "snapshot myzpool/filesystem@snap was created",
            "changes": {"myzpool/filesystem@snap": "snapshotted"},
        }

        mock_exists = MagicMock(return_value=False)
        mock_snapshot = MagicMock(return_value=OrderedDict([("snapshotted", True)]))
        with patch.dict(zfs.__salt__, {"zfs.exists": mock_exists}), patch.dict(
            zfs.__salt__, {"zfs.snapshot": mock_snapshot}
        ), patch.dict(zfs.__utils__, self.utils_patch):
            self.assertEqual(ret, zfs.snapshot_present("myzpool/filesystem@snap"))

    @slowTest
    def test_snapshot_present_fail(self):
        """
        Test if snapshot is present (using non existing snapshot)
        """
        ret = {
            "name": "myzpool/filesystem@snap",
            "result": False,
            "comment": "cannot open 'myzpool/filesystem': dataset does not exist",
            "changes": {},
        }

        mock_exists = MagicMock(return_value=False)
        mock_snapshot = MagicMock(
            return_value=OrderedDict(
                [
                    ("snapshotted", False),
                    (
                        "error",
                        "cannot open 'myzpool/filesystem': dataset does not exist",
                    ),
                ]
            )
        )
        with patch.dict(zfs.__salt__, {"zfs.exists": mock_exists}), patch.dict(
            zfs.__salt__, {"zfs.snapshot": mock_snapshot}
        ), patch.dict(zfs.__utils__, self.utils_patch):
            self.assertEqual(ret, zfs.snapshot_present("myzpool/filesystem@snap"))

    def test_propmoted(self):
        """
        Test promotion of clone (already promoted)
        """
        ret = {
            "name": "myzpool/filesystem",
            "result": True,
            "comment": "myzpool/filesystem already promoted",
            "changes": {},
        }

        mock_exists = MagicMock(return_value=True)
        mock_get = MagicMock(
            return_value=OrderedDict(
                [
                    (
                        "myzpool/filesystem",
                        OrderedDict([("origin", OrderedDict([("value", "-")]))]),
                    ),
                ]
            )
        )
        with patch.dict(zfs.__salt__, {"zfs.exists": mock_exists}), patch.dict(
            zfs.__salt__, {"zfs.get": mock_get}
        ), patch.dict(zfs.__utils__, self.utils_patch):
            self.assertEqual(ret, zfs.promoted("myzpool/filesystem"))

    @slowTest
    def test_propmoted_clone(self):
        """
        Test promotion of clone
        """
        ret = {
            "name": "myzpool/filesystem",
            "result": True,
            "comment": "myzpool/filesystem promoted",
            "changes": {"myzpool/filesystem": "promoted"},
        }

        mock_exists = MagicMock(return_value=True)
        mock_get = MagicMock(
            return_value=OrderedDict(
                [
                    (
                        "myzpool/filesystem",
                        OrderedDict(
                            [
                                (
                                    "origin",
                                    OrderedDict(
                                        [("value", "myzool/filesystem_source@clean")]
                                    ),
                                ),
                            ]
                        ),
                    ),
                ]
            )
        )
        mock_promote = MagicMock(return_value=OrderedDict([("promoted", True)]))
        with patch.dict(zfs.__salt__, {"zfs.exists": mock_exists}), patch.dict(
            zfs.__salt__, {"zfs.get": mock_get}
        ), patch.dict(zfs.__salt__, {"zfs.promote": mock_promote}), patch.dict(
            zfs.__utils__, self.utils_patch
        ):
            self.assertEqual(ret, zfs.promoted("myzpool/filesystem"))

    def test_propmoted_fail(self):
        """
        Test promotion of clone (unknown dataset)
        """
        ret = {
            "name": "myzpool/filesystem",
            "result": False,
            "comment": "dataset myzpool/filesystem does not exist",
            "changes": {},
        }

        mock_exists = MagicMock(return_value=False)
        with patch.dict(zfs.__salt__, {"zfs.exists": mock_exists}), patch.dict(
            zfs.__utils__, self.utils_patch
        ):
            self.assertEqual(ret, zfs.promoted("myzpool/filesystem"))

    @slowTest
    def test_scheduled_snapshot_fail(self):
        """
        Test scheduled_snapshot of unknown dataset
        """
        ret = {
            "name": "myzpool/filesystem",
            "result": False,
            "comment": "dataset myzpool/filesystem does not exist",
            "changes": {},
        }

        mock_exists = MagicMock(return_value=False)
        with patch.dict(zfs.__salt__, {"zfs.exists": mock_exists}), patch.dict(
            zfs.__utils__, self.utils_patch
        ):
            self.assertEqual(
                ret,
                zfs.scheduled_snapshot(
                    "myzpool/filesystem", "shadow", schedule={"hour": 6}
                ),
            )



# -*- coding: utf-8 -*-
# Import Python libs
from __future__ import absolute_import, print_function, unicode_literals

# Import Salt Libs
import salt.states.apache_conf as apache_conf

# Import Salt Testing Libs
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class ApacheConfTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.states.apache_conf
    """

    def setup_loader_modules(self):
        return {apache_conf: {}}

    # 'enabled' function tests: 1

    def test_enabled(self):
        """
        Test to ensure an Apache conf is enabled.
        """
        name = "saltstack.com"

        ret = {"name": name, "result": True, "changes": {}, "comment": ""}

        mock = MagicMock(side_effect=[True, False, False])
        mock_str = MagicMock(return_value={"Status": ["enabled"]})
        with patch.dict(
            apache_conf.__salt__,
            {"apache.check_conf_enabled": mock, "apache.a2enconf": mock_str},
        ):
            comt = "{0} already enabled.".format(name)
            ret.update({"comment": comt})
            self.assertDictEqual(apache_conf.enabled(name), ret)

            comt = "Apache conf {0} is set to be enabled.".format(name)
            ret.update(
                {"comment": comt, "result": None, "changes": {"new": name, "old": None}}
            )
            with patch.dict(apache_conf.__opts__, {"test": True}):
                self.assertDictEqual(apache_conf.enabled(name), ret)

            comt = "Failed to enable {0} Apache conf".format(name)
            ret.update({"comment": comt, "result": False, "changes": {}})
            with patch.dict(apache_conf.__opts__, {"test": False}):
                self.assertDictEqual(apache_conf.enabled(name), ret)

    # 'disabled' function tests: 1

    def test_disabled(self):
        """
        Test to ensure an Apache conf is disabled.
        """
        name = "saltstack.com"

        ret = {"name": name, "result": None, "changes": {}, "comment": ""}

        mock = MagicMock(side_effect=[True, True, False])
        mock_str = MagicMock(return_value={"Status": ["disabled"]})
        with patch.dict(
            apache_conf.__salt__,
            {"apache.check_conf_enabled": mock, "apache.a2disconf": mock_str},
        ):
            comt = "Apache conf {0} is set to be disabled.".format(name)
            ret.update({"comment": comt, "changes": {"new": None, "old": name}})
            with patch.dict(apache_conf.__opts__, {"test": True}):
                self.assertDictEqual(apache_conf.disabled(name), ret)

            comt = "Failed to disable {0} Apache conf".format(name)
            ret.update({"comment": comt, "result": False, "changes": {}})
            with patch.dict(apache_conf.__opts__, {"test": False}):
                self.assertDictEqual(apache_conf.disabled(name), ret)

            comt = "{0} already disabled.".format(name)
            ret.update({"comment": comt, "result": True})
            self.assertDictEqual(apache_conf.disabled(name), ret)



# -*- coding: utf-8 -*-

# Import future libs
from __future__ import absolute_import, print_function, unicode_literals

# Import 3rd-party libs
from io import BytesIO, StringIO

# Import salt module
import salt.modules.tomcat as tomcat
from salt.ext.six import string_types
from salt.ext.six.moves.urllib.request import (
    HTTPBasicAuthHandler as _HTTPBasicAuthHandler,
)
from salt.ext.six.moves.urllib.request import (
    HTTPDigestAuthHandler as _HTTPDigestAuthHandler,
)
from salt.ext.six.moves.urllib.request import build_opener as _build_opener

# Import Salt Testing libs
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class TomcatTestCasse(TestCase, LoaderModuleMockMixin):
    """
    Tests cases for salt.modules.tomcat
    """

    def setup_loader_modules(self):
        return {tomcat: {}}

    def test_tomcat_wget_no_bytestring(self):
        responses = {
            "string": StringIO("Best response ever\r\nAnd you know it!"),
            "bytes": BytesIO(b"Best response ever\r\nAnd you know it!"),
        }

        string_mock = MagicMock(return_value=responses["string"])
        bytes_mock = MagicMock(return_value=responses["bytes"])
        with patch(
            "salt.modules.tomcat._auth",
            MagicMock(
                return_value=_build_opener(
                    _HTTPBasicAuthHandler(), _HTTPDigestAuthHandler()
                )
            ),
        ):
            with patch("salt.modules.tomcat._urlopen", string_mock):
                response = tomcat._wget(
                    "tomcat.wait", url="http://localhost:8080/nofail"
                )
                for line in response["msg"]:
                    self.assertIsInstance(line, string_types)

            with patch("salt.modules.tomcat._urlopen", bytes_mock):
                try:
                    response = tomcat._wget(
                        "tomcat.wait", url="http://localhost:8080/nofail"
                    )
                except TypeError as type_error:
                    if (
                        type_error.args[0]
                        == "startswith first arg must be bytes or a tuple of bytes, not str"
                    ):
                        self.fail("Got back a byte string, should've been a string")
                    else:
                        raise type_error

                for line in response["msg"]:
                    self.assertIsInstance(line, string_types)



# -*- coding: utf-8 -*-
"""
:maintainer:    SaltStack
:maturity:      new
:platform:      all

Utilities supporting modules for Hashicorp Vault. Configuration instructions are
documented in the execution module docs.
"""

from __future__ import absolute_import, print_function, unicode_literals

import base64
import logging
import os
import time

import requests
import salt.crypt
import salt.exceptions
import salt.utils.versions

log = logging.getLogger(__name__)
logging.getLogger("requests").setLevel(logging.WARNING)


# Load the __salt__ dunder if not already loaded (when called from utils-module)
__salt__ = None


def __virtual__():  # pylint: disable=expected-2-blank-lines-found-0
    try:
        global __salt__  # pylint: disable=global-statement
        if not __salt__:
            __salt__ = salt.loader.minion_mods(__opts__)
            return True
    except Exception as e:  # pylint: disable=broad-except
        log.error("Could not load __salt__: %s", e)
        return False


def _get_token_and_url_from_master():
    """
    Get a token with correct policies for the minion, and the url to the Vault
    service
    """
    minion_id = __grains__["id"]
    pki_dir = __opts__["pki_dir"]
    # Allow minion override salt-master settings/defaults
    try:
        uses = __opts__.get("vault", {}).get("auth", {}).get("uses", None)
        ttl = __opts__.get("vault", {}).get("auth", {}).get("ttl", None)
    except (TypeError, AttributeError):
        # If uses or ttl are not defined, just use defaults
        uses = None
        ttl = None

    # When rendering pillars, the module executes on the master, but the token
    # should be issued for the minion, so that the correct policies are applied
    if __opts__.get("__role", "minion") == "minion":
        private_key = "{0}/minion.pem".format(pki_dir)
        log.debug("Running on minion, signing token request with key %s", private_key)
        signature = base64.b64encode(salt.crypt.sign_message(private_key, minion_id))
        result = __salt__["publish.runner"](
            "vault.generate_token", arg=[minion_id, signature, False, ttl, uses]
        )
    else:
        private_key = "{0}/master.pem".format(pki_dir)
        log.debug(
            "Running on master, signing token request for %s with key %s",
            minion_id,
            private_key,
        )
        signature = base64.b64encode(salt.crypt.sign_message(private_key, minion_id))
        result = __salt__["saltutil.runner"](
            "vault.generate_token",
            minion_id=minion_id,
            signature=signature,
            impersonated_by_master=True,
            ttl=ttl,
            uses=uses,
        )
    if not result:
        log.error(
            "Failed to get token from master! No result returned - "
            "is the peer publish configuration correct?"
        )
        raise salt.exceptions.CommandExecutionError(result)
    if not isinstance(result, dict):
        log.error(
            "Failed to get token from master! " "Response is not a dict: %s", result
        )
        raise salt.exceptions.CommandExecutionError(result)
    if "error" in result:
        log.error(
            "Failed to get token from master! " "An error was returned: %s",
            result["error"],
        )
        raise salt.exceptions.CommandExecutionError(result)
    if "session" in result.get("token_backend", "session"):
        # This is the only way that this key can be placed onto __context__
        # Thus is tells the minion that the master is configured for token_backend: session
        log.debug("Using session storage for vault credentials")
        __context__["vault_secret_path_metadata"] = {}
    return {
        "url": result["url"],
        "token": result["token"],
        "verify": result.get("verify", None),
        "uses": result.get("uses", 1),
        "lease_duration": result["lease_duration"],
        "issued": result["issued"],
    }


def get_vault_connection():
    """
    Get the connection details for calling Vault, from local configuration if
    it exists, or from the master otherwise
    """

    def _use_local_config():
        log.debug("Using Vault connection details from local config")
        try:
            if __opts__["vault"]["auth"]["method"] == "approle":
                verify = __opts__["vault"].get("verify", None)
                if _selftoken_expired():
                    log.debug("Vault token expired. Recreating one")
                    # Requesting a short ttl token
                    url = "{0}/v1/auth/approle/login".format(__opts__["vault"]["url"])
                    payload = {"role_id": __opts__["vault"]["auth"]["role_id"]}
                    if "secret_id" in __opts__["vault"]["auth"]:
                        payload["secret_id"] = __opts__["vault"]["auth"]["secret_id"]
                    response = requests.post(url, json=payload, verify=verify)
                    if response.status_code != 200:
                        errmsg = "An error occurred while getting a token from approle"
                        raise salt.exceptions.CommandExecutionError(errmsg)
                    __opts__["vault"]["auth"]["token"] = response.json()["auth"][
                        "client_token"
                    ]
            if __opts__["vault"]["auth"]["method"] == "wrapped_token":
                verify = __opts__["vault"].get("verify", None)
                if _wrapped_token_valid():
                    url = "{0}/v1/sys/wrapping/unwrap".format(__opts__["vault"]["url"])
                    headers = {"X-Vault-Token": __opts__["vault"]["auth"]["token"]}
                    response = requests.post(url, headers=headers, verify=verify)
                    if response.status_code != 200:
                        errmsg = "An error occured while unwrapping vault token"
                        raise salt.exceptions.CommandExecutionError(errmsg)
                    __opts__["vault"]["auth"]["token"] = response.json()["auth"][
                        "client_token"
                    ]
            return {
                "url": __opts__["vault"]["url"],
                "token": __opts__["vault"]["auth"]["token"],
                "verify": __opts__["vault"].get("verify", None),
                "issued": int(round(time.time())),
                "ttl": 3600,
            }
        except KeyError as err:
            errmsg = 'Minion has "vault" config section, but could not find key "{0}" within'.format(
                err
            )
            raise salt.exceptions.CommandExecutionError(errmsg)

    if "vault" in __opts__ and __opts__.get("__role", "minion") == "master":
        if "id" in __grains__:
            log.debug("Contacting master for Vault connection details")
            return _get_token_and_url_from_master()
        else:
            return _use_local_config()
    elif any(
        (
            __opts__.get("local", None),
            __opts__.get("file_client", None) == "local",
            __opts__.get("master_type", None) == "disable",
        )
    ):
        return _use_local_config()
    else:
        log.debug("Contacting master for Vault connection details")
        return _get_token_and_url_from_master()


def del_cache():
    """
    Delete cache file
    """
    log.debug("Deleting cache file")
    cache_file = os.path.join(__opts__["cachedir"], "salt_vault_token")

    if os.path.exists(cache_file):
        os.remove(cache_file)
    else:
        log.debug("Attempted to delete vault cache file, but it does not exist.")


def write_cache(connection):
    # If uses is 1 and unlimited_use_token is not true, then this is a single use token and should not be cached
    # In that case, we still want to cache the vault metadata lookup information for paths, so continue on
    if (
        connection.get("uses", None) == 1
        and "unlimited_use_token" not in connection
        and "vault_secret_path_metadata" not in connection
    ):
        log.debug("Not caching vault single use token")
        __context__["vault_token"] = connection
        return True
    elif (
        "vault_secret_path_metadata" in __context__
        and "vault_secret_path_metadata" not in connection
    ):
        # If session storage is being used, and info passed is not the already saved metadata
        log.debug("Storing token only for this session")
        __context__["vault_token"] = connection
        return True
    elif "vault_secret_path_metadata" in __context__:
        # Must have been passed metadata. This is already handled by _get_secret_path_metadata
        #  and does not need to be resaved
        return True

    cache_file = os.path.join(__opts__["cachedir"], "salt_vault_token")
    try:
        log.debug("Writing vault cache file")
        # Detect if token was issued without use limit
        if connection.get("uses") == 0:
            connection["unlimited_use_token"] = True
        else:
            connection["unlimited_use_token"] = False
        with salt.utils.files.fpopen(cache_file, "w", mode=0o600) as fp_:
            fp_.write(salt.utils.json.dumps(connection))
        return True
    except (IOError, OSError):
        log.error(
            "Failed to cache vault information", exc_info_on_loglevel=logging.DEBUG
        )
        return False


def _read_cache_file():
    """
    Return contents of cache file
    """
    try:
        cache_file = os.path.join(__opts__["cachedir"], "salt_vault_token")
        with salt.utils.files.fopen(cache_file, "r") as contents:
            return salt.utils.json.load(contents)
    except FileNotFoundError:
        return {}


def get_cache():
    """
    Return connection information from vault cache file
    """

    def _gen_new_connection():
        log.debug("Refreshing token")
        connection = get_vault_connection()
        write_status = write_cache(connection)
        return connection

    connection = _read_cache_file()
    # If no cache, or only metadata info is saved in cache, generate a new token
    if not connection or "url" not in connection:
        return _gen_new_connection()

    # Drop 10 seconds from ttl to be safe
    if "lease_duration" in connection:
        ttl = connection["lease_duration"]
    else:
        ttl = connection["ttl"]
    ttl10 = connection["issued"] + ttl - 10
    cur_time = int(round(time.time()))

    # Determine if ttl still valid
    if ttl10 < cur_time:
        log.debug("Cached token has expired {} < {}: DELETING".format(ttl10, cur_time))
        del_cache()
        return _gen_new_connection()
    else:
        log.debug("Token has not expired {} > {}".format(ttl10, cur_time))
    return connection


def make_request(
    method,
    resource,
    token=None,
    vault_url=None,
    get_token_url=False,
    retry=False,
    **args
):
    """
    Make a request to Vault
    """
    if "vault_token" in __context__:
        connection = __context__["vault_token"]
    else:
        connection = get_cache()
    token = connection["token"] if not token else token
    vault_url = connection["url"] if not vault_url else vault_url
    if "verify" in args:
        args["verify"] = args["verify"]
    else:
        try:
            args["verify"] = __opts__.get("vault").get("verify", None)
        except (TypeError, AttributeError):
            # Don't worry about setting verify if it doesn't exist
            pass
    url = "{0}/{1}".format(vault_url, resource)
    headers = {"X-Vault-Token": str(token), "Content-Type": "application/json"}
    response = requests.request(method, url, headers=headers, **args)
    if not response.ok and response.json().get("errors", None) == ["permission denied"]:
        log.info("Permission denied from vault")
        del_cache()
        if not retry:
            log.debug("Retrying with new credentials")
            response = make_request(
                method,
                resource,
                token=None,
                vault_url=vault_url,
                get_token_url=get_token_url,
                retry=True,
                **args
            )
        else:
            log.error("Unable to connect to vault server: %s", response.text)
            return response
    elif not response.ok:
        log.error("Error from vault: %s", response.text)
        return response

    # Decrement vault uses, only on secret URL lookups and multi use tokens
    if (
        not connection.get("unlimited_use_token")
        and not resource.startswith("v1/sys")
        and not resource.startswith("v1/sec")
    ):
        log.debug("Decrementing Vault uses on limited token for url: %s", resource)
        connection["uses"] -= 1
        if connection["uses"] <= 0:
            log.debug("Cached token has no more uses left.")
            if "vault_token" not in __context__:
                del_cache()
            else:
                log.debug("Deleting token from memory")
                del __context__["vault_token"]
        else:
            log.debug("Token has {} uses left".format(connection["uses"]))
            write_cache(connection)

    if get_token_url:
        return response, token, vault_url
    else:
        return response


def _selftoken_expired():
    """
    Validate the current token exists and is still valid
    """
    try:
        verify = __opts__["vault"].get("verify", None)
        url = "{0}/v1/auth/token/lookup-self".format(__opts__["vault"]["url"])
        if "token" not in __opts__["vault"]["auth"]:
            return True
        headers = {"X-Vault-Token": __opts__["vault"]["auth"]["token"]}
        response = requests.get(url, headers=headers, verify=verify)
        if response.status_code != 200:
            return True
        return False
    except Exception as e:  # pylint: disable=broad-except
        raise salt.exceptions.CommandExecutionError(
            "Error while looking up self token : {0}".format(e)
        )


def _wrapped_token_valid():
    """
    Validate the wrapped token exists and is still valid
    """
    try:
        verify = __opts__["vault"].get("verify", None)
        url = "{0}/v1/sys/wrapping/lookup".format(__opts__["vault"]["url"])
        if "token" not in __opts__["vault"]["auth"]:
            return False
        headers = {"X-Vault-Token": __opts__["vault"]["auth"]["token"]}
        response = requests.post(url, headers=headers, verify=verify)
        if response.status_code != 200:
            return False
        return True
    except Exception as e:  # pylint: disable=broad-except
        raise salt.exceptions.CommandExecutionError(
            "Error while looking up wrapped token : {0}".format(e)
        )


def is_v2(path):
    """
    Determines if a given secret path is kv version 1 or 2
    CLI Example:
    .. code-block:: bash
        salt '*' vault.is_v2 "secret/my/secret"
    """
    ret = {"v2": False, "data": path, "metadata": path, "delete": path, "type": None}
    path_metadata = _get_secret_path_metadata(path)
    if not path_metadata:
        # metadata lookup failed. Simply return not v2
        return ret
    ret["type"] = path_metadata.get("type", "kv")
    if ret["type"] == "kv" and path_metadata.get("options", {}).get("version", "1") in [
        "2"
    ]:
        ret["v2"] = True
        ret["data"] = _v2_the_path(path, path_metadata.get("path", path))
        ret["metadata"] = _v2_the_path(
            path, path_metadata.get("path", path), "metadata"
        )
        ret["destroy"] = _v2_the_path(path, path_metadata.get("path", path), "destroy")
    return ret


def _v2_the_path(path, pfilter, ptype="data"):
    """
    Given a path, a filter, and a path type, properly inject 'data' or 'metadata' into the path
    CLI Example:
    .. code-block:: python
        _v2_the_path('dev/secrets/fu/bar', 'dev/secrets', 'data') => 'dev/secrets/data/fu/bar'
    """
    possible_types = ["data", "metadata", "destroy"]
    assert ptype in possible_types
    msg = "Path {} already contains {} in the right place - saltstack duct tape?".format(
        path, ptype
    )

    path = path.rstrip("/").lstrip("/")
    pfilter = pfilter.rstrip("/").lstrip("/")

    together = pfilter + "/" + ptype

    otype = possible_types[0] if possible_types[0] != ptype else possible_types[1]
    other = pfilter + "/" + otype
    if path.startswith(other):
        path = path.replace(other, together, 1)
        msg = 'Path is a "{}" type but "{}" type requested - Flipping: {}'.format(
            otype, ptype, path
        )
    elif not path.startswith(together):
        msg = "Converting path to v2 {} => {}".format(
            path, path.replace(pfilter, together, 1)
        )
        path = path.replace(pfilter, together, 1)

    log.debug(msg)
    return path


def _get_secret_path_metadata(path):
    """
    Given a path, query vault to determine mount point, type, and version
    CLI Example:
    .. code-block:: python
        _get_secret_path_metadata('dev/secrets/fu/bar')
    """
    ckey = "vault_secret_path_metadata"

    # Attempt to lookup from cache
    if ckey in __context__:
        cache_content = __context__[ckey]
    else:
        cache_content = _read_cache_file()
    if ckey not in cache_content:
        cache_content[ckey] = {}

    ret = None
    if path.startswith(tuple(cache_content[ckey].keys())):
        log.debug("Found cached metadata for %s", path)
        ret = next(v for k, v in cache_content[ckey].items() if path.startswith(k))
    else:
        log.debug("Fetching metadata for %s", path)
        try:
            url = "v1/sys/internal/ui/mounts/{0}".format(path)
            response = make_request("GET", url)
            if response.ok:
                response.raise_for_status()
            if response.json().get("data", False):
                log.debug("Got metadata for %s", path)
                ret = response.json()["data"]
                # Write metadata to cache file
                # Check for new cache content from make_request
                if "url" not in cache_content:
                    if ckey in __context__:
                        cache_content = __context__[ckey]
                    else:
                        cache_content = _read_cache_file()
                    if ckey not in cache_content:
                        cache_content[ckey] = {}
                cache_content[ckey][path] = ret
                write_cache(cache_content)
            else:
                raise response.json()
        except Exception as err:  # pylint: disable=broad-except
            log.error("Failed to get secret metadata %s: %s", type(err).__name__, err)
    return ret



# -*- coding: utf-8 -*-
"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""
# Import Python libs
from __future__ import absolute_import, print_function, unicode_literals

# Import Salt Libs
import salt.states.boto_cloudwatch_alarm as boto_cloudwatch_alarm

# Import Salt Testing Libs
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class BotoCloudwatchAlarmTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.states.boto_cloudwatch_alarm
    """

    def setup_loader_modules(self):
        return {boto_cloudwatch_alarm: {}}

    # 'present' function tests: 1

    def test_present(self):
        """
        Test to ensure the cloudwatch alarm exists.
        """
        name = "my test alarm"
        attributes = {
            "metric": "ApproximateNumberOfMessagesVisible",
            "namespace": "AWS/SQS",
        }

        ret = {"name": name, "result": None, "changes": {}, "comment": ""}

        mock = MagicMock(side_effect=[["ok_actions"], [], []])
        mock_bool = MagicMock(return_value=True)
        with patch.dict(
            boto_cloudwatch_alarm.__salt__,
            {
                "boto_cloudwatch.get_alarm": mock,
                "boto_cloudwatch.create_or_update_alarm": mock_bool,
            },
        ):
            with patch.dict(boto_cloudwatch_alarm.__opts__, {"test": True}):
                comt = "alarm my test alarm is to be created/updated."
                ret.update({"comment": comt})
                self.assertDictEqual(
                    boto_cloudwatch_alarm.present(name, attributes), ret
                )

                comt = "alarm my test alarm is to be created/updated."
                ret.update({"comment": comt})
                self.assertDictEqual(
                    boto_cloudwatch_alarm.present(name, attributes), ret
                )

            with patch.dict(boto_cloudwatch_alarm.__opts__, {"test": False}):
                changes = {
                    "new": {
                        "metric": "ApproximateNumberOfMessagesVisible",
                        "namespace": "AWS/SQS",
                    }
                }
                comt = "alarm my test alarm is to be created/updated."
                ret.update({"changes": changes, "comment": "", "result": True})
                self.assertDictEqual(
                    boto_cloudwatch_alarm.present(name, attributes), ret
                )

    # 'absent' function tests: 1

    def test_absent(self):
        """
        Test to ensure the named cloudwatch alarm is deleted.
        """
        name = "my test alarm"

        ret = {"name": name, "result": None, "changes": {}, "comment": ""}

        mock = MagicMock(side_effect=[True, False])
        with patch.dict(
            boto_cloudwatch_alarm.__salt__, {"boto_cloudwatch.get_alarm": mock}
        ):
            with patch.dict(boto_cloudwatch_alarm.__opts__, {"test": True}):
                comt = "alarm {0} is set to be removed.".format(name)
                ret.update({"comment": comt})
                self.assertDictEqual(boto_cloudwatch_alarm.absent(name), ret)

                comt = "my test alarm does not exist in None."
                ret.update({"comment": comt, "result": True})
                self.assertDictEqual(boto_cloudwatch_alarm.absent(name), ret)



# -*- coding: utf-8 -*-
"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""
# Import Python libs
from __future__ import absolute_import, print_function, unicode_literals

# Import Salt Libs
import salt.states.rbenv as rbenv

# Import Salt Testing Libs
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class RbenvTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.states.rbenv
    """

    def setup_loader_modules(self):
        return {rbenv: {}}

    # 'installed' function tests: 1

    def test_installed(self):
        """
        Test to verify that the specified ruby is installed with rbenv.
        """
        # rbenv.is_installed is used wherever test is False.
        mock_is = MagicMock(side_effect=[False, True, True, True, True])

        # rbenv.install is only called when an action is attempted
        # (ie. Successfully... or Failed...)
        mock_i = MagicMock(side_effect=[False, False, False])

        # rbenv.install_ruby is only called when rbenv is successfully
        # installed and an attempt to install a version of Ruby is
        # made.
        mock_ir = MagicMock(side_effect=[True, False])
        mock_def = MagicMock(return_value="2.3.4")
        mock_ver = MagicMock(return_value=["2.3.4", "2.4.1"])
        with patch.dict(
            rbenv.__salt__,
            {
                "rbenv.is_installed": mock_is,
                "rbenv.install": mock_i,
                "rbenv.default": mock_def,
                "rbenv.versions": mock_ver,
                "rbenv.install_ruby": mock_ir,
            },
        ):
            with patch.dict(rbenv.__opts__, {"test": True}):
                name = "1.9.3-p551"
                comt = "Ruby {0} is set to be installed".format(name)
                ret = {"name": name, "changes": {}, "comment": comt, "result": None}
                self.assertDictEqual(rbenv.installed(name), ret)

                name = "2.4.1"
                comt = "Ruby {0} is already installed".format(name)
                ret = {
                    "name": name,
                    "changes": {},
                    "comment": comt,
                    "default": False,
                    "result": True,
                }
                self.assertDictEqual(rbenv.installed(name), ret)

                name = "2.3.4"
                comt = "Ruby {0} is already installed".format(name)
                ret = {
                    "name": name,
                    "changes": {},
                    "comment": comt,
                    "default": True,
                    "result": True,
                }
                self.assertDictEqual(rbenv.installed(name), ret)

            with patch.dict(rbenv.__opts__, {"test": False}):
                name = "2.4.1"
                comt = "Rbenv failed to install"
                ret = {"name": name, "changes": {}, "comment": comt, "result": False}
                self.assertDictEqual(rbenv.installed(name), ret)

                comt = "Requested ruby exists"
                ret = {
                    "name": name,
                    "comment": comt,
                    "default": False,
                    "changes": {},
                    "result": True,
                }
                self.assertDictEqual(rbenv.installed(name), ret)

                name = "2.3.4"
                comt = "Requested ruby exists"
                ret = {
                    "name": name,
                    "comment": comt,
                    "default": True,
                    "changes": {},
                    "result": True,
                }
                self.assertDictEqual(rbenv.installed(name), ret)

                name = "1.9.3-p551"
                comt = "Successfully installed ruby"
                ret = {
                    "name": name,
                    "comment": comt,
                    "default": False,
                    "changes": {name: "Installed"},
                    "result": True,
                }
                self.assertDictEqual(rbenv.installed(name), ret)

                comt = "Failed to install ruby"
                ret = {"name": name, "comment": comt, "changes": {}, "result": False}
                self.assertDictEqual(rbenv.installed(name), ret)

    # 'absent' function tests: 1

    def test_absent(self):
        """
        Test to verify that the specified ruby is not installed with rbenv.
        """
        # rbenv.is_installed is used for all tests here.
        mock_is = MagicMock(
            side_effect=[False, True, True, True, False, True, True, True, True, True]
        )
        # rbenv.uninstall_ruby is only called when an action is
        # attempted (ie. Successfully... or Failed...)
        mock_uninstalled = MagicMock(side_effect=[True, False, False, True])
        mock_def = MagicMock(return_value="2.3.4")
        mock_ver = MagicMock(return_value=["2.3.4", "2.4.1"])
        with patch.dict(
            rbenv.__salt__,
            {
                "rbenv.is_installed": mock_is,
                "rbenv.default": mock_def,
                "rbenv.versions": mock_ver,
                "rbenv.uninstall_ruby": mock_uninstalled,
            },
        ):

            with patch.dict(rbenv.__opts__, {"test": True}):
                name = "1.9.3-p551"
                comt = "Rbenv not installed, {0} not either".format(name)
                ret = {"name": name, "changes": {}, "comment": comt, "result": True}
                self.assertDictEqual(rbenv.absent(name), ret)

                comt = "Ruby {0} is already uninstalled".format(name)
                ret = {"name": name, "changes": {}, "comment": comt, "result": True}
                self.assertDictEqual(rbenv.absent(name), ret)

                name = "2.3.4"
                comt = "Ruby {0} is set to be uninstalled".format(name)
                ret = {
                    "name": name,
                    "changes": {},
                    "comment": comt,
                    "default": True,
                    "result": None,
                }
                self.assertDictEqual(rbenv.absent("2.3.4"), ret)

                name = "2.4.1"
                comt = "Ruby {0} is set to be uninstalled".format(name)
                ret = {
                    "name": name,
                    "changes": {},
                    "comment": comt,
                    "default": False,
                    "result": None,
                }
                self.assertDictEqual(rbenv.absent("2.4.1"), ret)

            with patch.dict(rbenv.__opts__, {"test": False}):
                name = "1.9.3-p551"
                comt = "Rbenv not installed, {0} not either".format(name)
                ret = {"name": name, "changes": {}, "comment": comt, "result": True}
                self.assertDictEqual(rbenv.absent(name), ret)

                comt = "Ruby {0} is already absent".format(name)
                ret = {"name": name, "changes": {}, "comment": comt, "result": True}
                self.assertDictEqual(rbenv.absent(name), ret)

                name = "2.3.4"
                comt = "Successfully removed ruby"
                ret = {
                    "name": name,
                    "changes": {name: "Uninstalled"},
                    "comment": comt,
                    "default": True,
                    "result": True,
                }
                self.assertDictEqual(rbenv.absent(name), ret)

                comt = "Failed to uninstall ruby"
                ret = {
                    "name": name,
                    "changes": {},
                    "comment": comt,
                    "default": True,
                    "result": False,
                }
                self.assertDictEqual(rbenv.absent(name), ret)

                name = "2.4.1"
                comt = "Failed to uninstall ruby"
                ret = {
                    "name": name,
                    "changes": {},
                    "comment": comt,
                    "default": False,
                    "result": False,
                }
                self.assertDictEqual(rbenv.absent(name), ret)

                comt = "Successfully removed ruby"
                ret = {
                    "name": name,
                    "changes": {name: "Uninstalled"},
                    "comment": comt,
                    "default": False,
                    "result": True,
                }
                self.assertDictEqual(rbenv.absent(name), ret)

    # 'install_rbenv' function tests: 1

    def test_install_rbenv(self):
        """
        Test to install rbenv if not installed.
        """
        name = "myqueue"

        ret = {"name": name, "changes": {}, "result": True, "comment": ""}

        mock_is = MagicMock(side_effect=[False, True, True, False, False])
        mock_i = MagicMock(side_effect=[False, True])
        with patch.dict(
            rbenv.__salt__, {"rbenv.is_installed": mock_is, "rbenv.install": mock_i}
        ):

            with patch.dict(rbenv.__opts__, {"test": True}):
                comt = "Rbenv is set to be installed"
                ret.update({"comment": comt, "result": None})
                self.assertDictEqual(rbenv.install_rbenv(name), ret)

                comt = "Rbenv is already installed"
                ret.update({"comment": comt, "result": True})
                self.assertDictEqual(rbenv.install_rbenv(name), ret)

            with patch.dict(rbenv.__opts__, {"test": False}):
                comt = "Rbenv is already installed"
                ret.update({"comment": comt, "result": True})
                self.assertDictEqual(rbenv.install_rbenv(name), ret)

                comt = "Rbenv failed to install"
                ret.update({"comment": comt, "result": False})
                self.assertDictEqual(rbenv.install_rbenv(name), ret)

                comt = "Rbenv installed"
                ret.update({"comment": comt, "result": True})
                self.assertDictEqual(rbenv.install_rbenv(name), ret)



# -*- coding: utf-8 -*-
"""
Functions which implement running reactor jobs
"""


# Import python libs
from __future__ import absolute_import, print_function, unicode_literals

import fnmatch
import glob
import logging
import os

# Import salt libs
import salt.client
import salt.defaults.exitcodes
import salt.runner
import salt.state
import salt.utils.args
import salt.utils.cache
import salt.utils.data
import salt.utils.event
import salt.utils.files
import salt.utils.master
import salt.utils.process
import salt.utils.yaml
import salt.wheel

# Import 3rd-party libs
from salt.ext import six

log = logging.getLogger(__name__)

REACTOR_INTERNAL_KEYWORDS = frozenset(
    ["__id__", "__sls__", "name", "order", "fun", "key", "state"]
)


class Reactor(salt.utils.process.SignalHandlingProcess, salt.state.Compiler):
    """
    Read in the reactor configuration variable and compare it to events
    processed on the master.
    The reactor has the capability to execute pre-programmed executions
    as reactions to events
    """

    aliases = {
        "cmd": "local",
    }

    def __init__(self, opts, **kwargs):
        super(Reactor, self).__init__(**kwargs)
        local_minion_opts = opts.copy()
        local_minion_opts["file_client"] = "local"
        self.minion = salt.minion.MasterMinion(local_minion_opts)
        salt.state.Compiler.__init__(self, opts, self.minion.rend)
        self.is_leader = True

    # We need __setstate__ and __getstate__ to avoid pickling errors since
    # 'self.rend' (from salt.state.Compiler) contains a function reference
    # which is not picklable.
    # These methods are only used when pickling so will not be used on
    # non-Windows platforms.
    def __setstate__(self, state):
        Reactor.__init__(
            self,
            state["opts"],
            log_queue=state["log_queue"],
            log_queue_level=state["log_queue_level"],
        )

    def __getstate__(self):
        return {
            "opts": self.opts,
            "log_queue": self.log_queue,
            "log_queue_level": self.log_queue_level,
        }

    def render_reaction(self, glob_ref, tag, data):
        """
        Execute the render system against a single reaction file and return
        the data structure
        """
        react = {}

        if glob_ref.startswith("salt://"):
            glob_ref = self.minion.functions["cp.cache_file"](glob_ref) or ""
        globbed_ref = glob.glob(glob_ref)
        if not globbed_ref:
            log.error(
                "Can not render SLS %s for tag %s. File missing or not found.",
                glob_ref,
                tag,
            )
        for fn_ in globbed_ref:
            try:
                res = self.render_template(fn_, tag=tag, data=data)

                # for #20841, inject the sls name here since verify_high()
                # assumes it exists in case there are any errors
                for name in res:
                    res[name]["__sls__"] = fn_

                react.update(res)
            except Exception:  # pylint: disable=broad-except
                log.exception('Failed to render "%s": ', fn_)
        return react

    def list_reactors(self, tag):
        """
        Take in the tag from an event and return a list of the reactors to
        process
        """
        log.debug("Gathering reactors for tag %s", tag)
        reactors = []
        if isinstance(self.opts["reactor"], six.string_types):
            try:
                with salt.utils.files.fopen(self.opts["reactor"]) as fp_:
                    react_map = salt.utils.yaml.safe_load(fp_)
            except (OSError, IOError):
                log.error('Failed to read reactor map: "%s"', self.opts["reactor"])
            except Exception:  # pylint: disable=broad-except
                log.error(
                    'Failed to parse YAML in reactor map: "%s"', self.opts["reactor"]
                )
        else:
            react_map = self.opts["reactor"]
        for ropt in react_map:
            if not isinstance(ropt, dict):
                continue
            if len(ropt) != 1:
                continue
            key = next(six.iterkeys(ropt))
            val = ropt[key]
            if fnmatch.fnmatch(tag, key):
                if isinstance(val, six.string_types):
                    reactors.append(val)
                elif isinstance(val, list):
                    reactors.extend(val)
        return reactors

    def list_all(self):
        """
        Return a list of the reactors
        """
        if isinstance(self.minion.opts["reactor"], six.string_types):
            log.debug("Reading reactors from yaml %s", self.opts["reactor"])
            try:
                with salt.utils.files.fopen(self.opts["reactor"]) as fp_:
                    react_map = salt.utils.yaml.safe_load(fp_)
            except (OSError, IOError):
                log.error('Failed to read reactor map: "%s"', self.opts["reactor"])
            except Exception:  # pylint: disable=broad-except
                log.error(
                    'Failed to parse YAML in reactor map: "%s"', self.opts["reactor"]
                )
        else:
            log.debug("Not reading reactors from yaml")
            react_map = self.minion.opts["reactor"]
        return react_map

    def add_reactor(self, tag, reaction):
        """
        Add a reactor
        """
        reactors = self.list_all()
        for reactor in reactors:
            _tag = next(six.iterkeys(reactor))
            if _tag == tag:
                return {"status": False, "comment": "Reactor already exists."}

        self.minion.opts["reactor"].append({tag: reaction})
        return {"status": True, "comment": "Reactor added."}

    def delete_reactor(self, tag):
        """
        Delete a reactor
        """
        reactors = self.list_all()
        for reactor in reactors:
            _tag = next(six.iterkeys(reactor))
            if _tag == tag:
                self.minion.opts["reactor"].remove(reactor)
                return {"status": True, "comment": "Reactor deleted."}

        return {"status": False, "comment": "Reactor does not exists."}

    def resolve_aliases(self, chunks):
        """
        Preserve backward compatibility by rewriting the 'state' key in the low
        chunks if it is using a legacy type.
        """
        for idx, _ in enumerate(chunks):
            new_state = self.aliases.get(chunks[idx]["state"])
            if new_state is not None:
                chunks[idx]["state"] = new_state

    def reactions(self, tag, data, reactors):
        """
        Render a list of reactor files and returns a reaction struct
        """
        log.debug("Compiling reactions for tag %s", tag)
        high = {}
        chunks = []
        try:
            for fn_ in reactors:
                high.update(self.render_reaction(fn_, tag, data))
            if high:
                errors = self.verify_high(high)
                if errors:
                    log.error(
                        "Unable to render reactions for event %s due to "
                        "errors (%s) in one or more of the sls files (%s)",
                        tag,
                        errors,
                        reactors,
                    )
                    return []  # We'll return nothing since there was an error
                chunks = self.order_chunks(self.compile_high_data(high))
        except Exception as exc:  # pylint: disable=broad-except
            log.exception("Exception encountered while compiling reactions")

        self.resolve_aliases(chunks)
        return chunks

    def call_reactions(self, chunks):
        """
        Execute the reaction state
        """
        for chunk in chunks:
            self.wrap.run(chunk)

    def run(self):
        """
        Enter into the server loop
        """
        salt.utils.process.appendproctitle(self.__class__.__name__)

        if self.opts["reactor_niceness"] and not salt.utils.platform.is_windows():
            log.info("Reactor setting niceness to %i", self.opts["reactor_niceness"])
            os.nice(self.opts["reactor_niceness"])

        # instantiate some classes inside our new process
        with salt.utils.event.get_event(
            self.opts["__role"],
            self.opts["sock_dir"],
            self.opts["transport"],
            opts=self.opts,
            listen=True,
        ) as event:
            self.wrap = ReactWrap(self.opts)

            for data in event.iter_events(full=True):
                # skip all events fired by ourselves
                if data["data"].get("user") == self.wrap.event_user:
                    continue

                # NOTE: these events must contain the masters key in order to be accepted
                # see salt.runners.reactor for the requesting interface
                if "salt/reactors/manage" in data["tag"]:
                    master_key = salt.utils.master.get_master_key("root", self.opts)
                    if data["data"].get("key") != master_key:
                        log.error(
                            "received salt/reactors/manage event without matching master_key. discarding"
                        )
                        continue
                if data["tag"].endswith("salt/reactors/manage/is_leader"):
                    event.fire_event(
                        {"result": self.is_leader}, "salt/reactors/manage/leader/value"
                    )
                if data["tag"].endswith("salt/reactors/manage/set_leader"):
                    # we only want to register events from the local master
                    if data["data"].get("id") == self.opts["id"]:
                        self.is_leader = data["data"]["value"]
                    event.fire_event(
                        {"result": self.is_leader}, "salt/reactors/manage/leader/value"
                    )
                if data["tag"].endswith("salt/reactors/manage/add"):
                    _data = data["data"]
                    res = self.add_reactor(_data["event"], _data["reactors"])
                    event.fire_event(
                        {"reactors": self.list_all(), "result": res},
                        "salt/reactors/manage/add-complete",
                    )
                elif data["tag"].endswith("salt/reactors/manage/delete"):
                    _data = data["data"]
                    res = self.delete_reactor(_data["event"])
                    event.fire_event(
                        {"reactors": self.list_all(), "result": res},
                        "salt/reactors/manage/delete-complete",
                    )
                elif data["tag"].endswith("salt/reactors/manage/list"):
                    event.fire_event(
                        {"reactors": self.list_all()},
                        "salt/reactors/manage/list-results",
                    )
                else:
                    # do not handle any reactions if not leader in cluster
                    if not self.is_leader:
                        continue
                    else:
                        reactors = self.list_reactors(data["tag"])
                        if not reactors:
                            continue
                        chunks = self.reactions(data["tag"], data["data"], reactors)
                        if chunks:
                            try:
                                self.call_reactions(chunks)
                            except SystemExit:
                                log.warning("Exit ignored by reactor")


class ReactWrap(object):
    """
    Wrapper that executes low data for the Reactor System
    """

    # class-wide cache of clients
    client_cache = None
    event_user = "Reactor"

    reaction_class = {
        "local": salt.client.LocalClient,
        "runner": salt.runner.RunnerClient,
        "wheel": salt.wheel.Wheel,
        "caller": salt.client.Caller,
    }

    def __init__(self, opts):
        self.opts = opts
        if ReactWrap.client_cache is None:
            ReactWrap.client_cache = salt.utils.cache.CacheDict(
                opts["reactor_refresh_interval"]
            )

        self.pool = salt.utils.process.ThreadPool(
            self.opts["reactor_worker_threads"],  # number of workers for runner/wheel
            queue_size=self.opts["reactor_worker_hwm"],  # queue size for those workers
        )

    def populate_client_cache(self, low):
        """
        Populate the client cache with an instance of the specified type
        """
        reaction_type = low["state"]
        if reaction_type not in self.client_cache:
            log.debug("Reactor is populating %s client cache", reaction_type)
            if reaction_type in ("runner", "wheel"):
                # Reaction types that run locally on the master want the full
                # opts passed.
                self.client_cache[reaction_type] = self.reaction_class[reaction_type](
                    self.opts
                )
                # The len() function will cause the module functions to load if
                # they aren't already loaded. We want to load them so that the
                # spawned threads don't need to load them. Loading in the
                # spawned threads creates race conditions such as sometimes not
                # finding the required function because another thread is in
                # the middle of loading the functions.
                len(self.client_cache[reaction_type].functions)
            else:
                # Reactions which use remote pubs only need the conf file when
                # instantiating a client instance.
                self.client_cache[reaction_type] = self.reaction_class[reaction_type](
                    self.opts["conf_file"]
                )

    def run(self, low):
        """
        Execute a reaction by invoking the proper wrapper func
        """
        self.populate_client_cache(low)
        try:
            l_fun = getattr(self, low["state"])
        except AttributeError:
            log.error("ReactWrap is missing a wrapper function for '%s'", low["state"])

        try:
            wrap_call = salt.utils.args.format_call(l_fun, low)
            args = wrap_call.get("args", ())
            kwargs = wrap_call.get("kwargs", {})
            # TODO: Setting user doesn't seem to work for actual remote pubs
            if low["state"] in ("runner", "wheel"):
                # Update called function's low data with event user to
                # segregate events fired by reactor and avoid reaction loops
                kwargs["__user__"] = self.event_user
                # Replace ``state`` kwarg which comes from high data compiler.
                # It breaks some runner functions and seems unnecessary.
                kwargs["__state__"] = kwargs.pop("state")
                # NOTE: if any additional keys are added here, they will also
                # need to be added to filter_kwargs()

            if "args" in kwargs:
                # New configuration
                reactor_args = kwargs.pop("args")
                for item in ("arg", "kwarg"):
                    if item in low:
                        log.warning(
                            "Reactor '%s' is ignoring '%s' param %s due to "
                            "presence of 'args' param. Check the Reactor System "
                            "documentation for the correct argument format.",
                            low["__id__"],
                            item,
                            low[item],
                        )
                if (
                    low["state"] == "caller"
                    and isinstance(reactor_args, list)
                    and not salt.utils.data.is_dictlist(reactor_args)
                ):
                    # Legacy 'caller' reactors were already using the 'args'
                    # param, but only supported a list of positional arguments.
                    # If low['args'] is a list but is *not* a dictlist, then
                    # this is actually using the legacy configuration. So, put
                    # the reactor args into kwarg['arg'] so that the wrapper
                    # interprets them as positional args.
                    kwargs["arg"] = reactor_args
                    kwargs["kwarg"] = {}
                else:
                    kwargs["arg"] = ()
                    kwargs["kwarg"] = reactor_args
                if not isinstance(kwargs["kwarg"], dict):
                    kwargs["kwarg"] = salt.utils.data.repack_dictlist(kwargs["kwarg"])
                    if not kwargs["kwarg"]:
                        log.error(
                            "Reactor '%s' failed to execute %s '%s': "
                            "Incorrect argument format, check the Reactor System "
                            "documentation for the correct format.",
                            low["__id__"],
                            low["state"],
                            low["fun"],
                        )
                        return
            else:
                # Legacy configuration
                react_call = {}
                if low["state"] in ("runner", "wheel"):
                    if "arg" not in kwargs or "kwarg" not in kwargs:
                        # Runner/wheel execute on the master, so we can use
                        # format_call to get the functions args/kwargs
                        react_fun = self.client_cache[low["state"]].functions.get(
                            low["fun"]
                        )
                        if react_fun is None:
                            log.error(
                                "Reactor '%s' failed to execute %s '%s': "
                                "function not available",
                                low["__id__"],
                                low["state"],
                                low["fun"],
                            )
                            return

                        react_call = salt.utils.args.format_call(
                            react_fun, low, expected_extra_kws=REACTOR_INTERNAL_KEYWORDS
                        )

                if "arg" not in kwargs:
                    kwargs["arg"] = react_call.get("args", ())
                if "kwarg" not in kwargs:
                    kwargs["kwarg"] = react_call.get("kwargs", {})

            # Execute the wrapper with the proper args/kwargs. kwargs['arg']
            # and kwargs['kwarg'] contain the positional and keyword arguments
            # that will be passed to the client interface to execute the
            # desired runner/wheel/remote-exec/etc. function.
            ret = l_fun(*args, **kwargs)

            if ret is False:
                log.error(
                    "Reactor '%s' failed  to execute %s '%s': "
                    "TaskPool queue is full!"
                    "Consider tuning reactor_worker_threads and/or"
                    " reactor_worker_hwm",
                    low["__id__"],
                    low["state"],
                    low["fun"],
                )

        except SystemExit:
            log.warning("Reactor '%s' attempted to exit. Ignored.", low["__id__"])
        except Exception:  # pylint: disable=broad-except
            log.error(
                "Reactor '%s' failed to execute %s '%s'",
                low["__id__"],
                low["state"],
                low["fun"],
                exc_info=True,
            )

    def runner(self, fun, **kwargs):
        """
        Wrap RunnerClient for executing :ref:`runner modules <all-salt.runners>`
        """
        return self.pool.fire_async(self.client_cache["runner"].low, args=(fun, kwargs))

    def wheel(self, fun, **kwargs):
        """
        Wrap Wheel to enable executing :ref:`wheel modules <all-salt.wheel>`
        """
        return self.pool.fire_async(self.client_cache["wheel"].low, args=(fun, kwargs))

    def local(self, fun, tgt, **kwargs):
        """
        Wrap LocalClient for running :ref:`execution modules <all-salt.modules>`
        """
        self.client_cache["local"].cmd_async(tgt, fun, **kwargs)

    def caller(self, fun, **kwargs):
        """
        Wrap LocalCaller to execute remote exec functions locally on the Minion
        """
        self.client_cache["caller"].cmd(fun, *kwargs["arg"], **kwargs["kwarg"])



# -*- coding: utf-8 -*-
"""
    :codeauthor: Rahul Handay <rahulha@saltstack.com>
"""

# Import Python Libs
from __future__ import absolute_import, print_function, unicode_literals

# Import Salt Libs
import salt.states.vbox_guest as vbox_guest

# Import Salt Testing Libs
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class VboxGuestTestCase(TestCase, LoaderModuleMockMixin):
    """
        Validate the vbox_guest state
    """

    def setup_loader_modules(self):
        return {vbox_guest: {}}

    def test_additions_installed(self):
        """
            Test to ensure that the VirtualBox Guest Additions are installed
        """
        ret = {"name": "salt", "changes": {}, "result": True, "comment": ""}
        mock = MagicMock(side_effect=[True, False, False, False])
        with patch.dict(
            vbox_guest.__salt__,
            {
                "vbox_guest.additions_version": mock,
                "vbox_guest.additions_install": mock,
            },
        ):
            ret.update({"comment": "System already in the correct state"})
            self.assertDictEqual(vbox_guest.additions_installed("salt"), ret)

            with patch.dict(vbox_guest.__opts__, {"test": True}):
                ret.update(
                    {
                        "changes": {"new": True, "old": False},
                        "comment": "The state of VirtualBox Guest"
                        " Additions will be changed.",
                        "result": None,
                    }
                )
                self.assertDictEqual(vbox_guest.additions_installed("salt"), ret)

            with patch.dict(vbox_guest.__opts__, {"test": False}):
                ret.update(
                    {
                        "changes": {"new": False, "old": False},
                        "comment": "The state of VirtualBox Guest"
                        " Additions was changed!",
                        "result": False,
                    }
                )
                self.assertDictEqual(vbox_guest.additions_installed("salt"), ret)

    def test_additions_removed(self):
        """
            Test to ensure that the VirtualBox Guest Additions are removed.
        """
        ret = {"name": "salt", "changes": {}, "result": True, "comment": ""}
        mock = MagicMock(side_effect=[False, True, True, True])
        with patch.dict(
            vbox_guest.__salt__,
            {"vbox_guest.additions_version": mock, "vbox_guest.additions_remove": mock},
        ):
            ret.update({"comment": "System already in the correct state"})
            self.assertDictEqual(vbox_guest.additions_removed("salt"), ret)

            with patch.dict(vbox_guest.__opts__, {"test": True}):
                ret.update(
                    {
                        "changes": {"new": True, "old": True},
                        "comment": "The state of VirtualBox Guest"
                        " Additions will be changed.",
                        "result": None,
                    }
                )
                self.assertDictEqual(vbox_guest.additions_removed("salt"), ret)

            with patch.dict(vbox_guest.__opts__, {"test": False}):
                ret.update(
                    {
                        "comment": "The state of VirtualBox Guest"
                        " Additions was changed!",
                        "result": True,
                    }
                )
                self.assertDictEqual(vbox_guest.additions_removed("salt"), ret)

    def test_grantaccess_to_sharedfolders(self):
        """
            Test to grant access to auto-mounted shared folders to the users.
        """
        ret = {"name": "AB", "changes": {}, "result": True, "comment": ""}
        mock = MagicMock(side_effect=[["AB"], "salt", "salt", "salt"])
        with patch.dict(
            vbox_guest.__salt__,
            {
                "vbox_guest.list_shared_folders_users": mock,
                "vbox_guest.grant_access_to_shared_folders_to": mock,
            },
        ):
            ret.update({"comment": "System already in the correct state"})
            self.assert_method(ret)

            with patch.dict(vbox_guest.__opts__, {"test": True}):
                ret.update(
                    {
                        "changes": {"new": ["AB"], "old": "salt"},
                        "comment": "List of users who have access to"
                        " auto-mounted shared folders will be changed",
                        "result": None,
                    }
                )
                self.assert_method(ret)

            with patch.dict(vbox_guest.__opts__, {"test": False}):
                ret.update(
                    {
                        "changes": {"new": "salt", "old": "salt"},
                        "comment": "List of users who have access to"
                        " auto-mounted shared folders was changed",
                        "result": True,
                    }
                )
                self.assert_method(ret)

    def assert_method(self, ret):
        """
            Method call for assert statements
        """
        self.assertDictEqual(vbox_guest.grant_access_to_shared_folders_to("AB"), ret)



# -*- coding: utf-8 -*-

# Import Python Libs
from __future__ import absolute_import, print_function, unicode_literals

# Import Salt libs
import salt.renderers.nacl as nacl

# Import Salt Testing libs
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class NaclTestCase(TestCase, LoaderModuleMockMixin):
    """
    unit test NaCl renderer
    """

    def setup_loader_modules(self):
        return {nacl: {}}

    def test__decrypt_object(self):
        """
        test _decrypt_object
        """
        secret = "Use more salt."
        crypted = "NACL[MRN3cc+fmdxyQbz6WMF+jq1hKdU5X5BBI7OjK+atvHo1ll+w1gZ7XyWtZVfq9gK9rQaMfkDxmidJKwE0Mw==]"

        secret_map = {"secret": secret}
        crypted_map = {"secret": crypted}

        secret_list = [secret]
        crypted_list = [crypted]

        with patch.dict(nacl.__salt__, {"nacl.dec": MagicMock(return_value=secret)}):
            self.assertEqual(nacl._decrypt_object(secret), secret)
            self.assertEqual(nacl._decrypt_object(crypted), secret)
            self.assertEqual(nacl._decrypt_object(crypted_map), secret_map)
            self.assertEqual(nacl._decrypt_object(crypted_list), secret_list)
            self.assertEqual(nacl._decrypt_object(None), None)

    def test_render(self):
        """
        test render
        """
        secret = "Use more salt."
        crypted = "NACL[MRN3cc+fmdxyQbz6WMF+jq1hKdU5X5BBI7OjK+atvHo1ll+w1gZ7XyWtZVfq9gK9rQaMfkDxmidJKwE0Mw==]"
        with patch.dict(nacl.__salt__, {"nacl.dec": MagicMock(return_value=secret)}):
            self.assertEqual(nacl.render(crypted), secret)



# -*- coding: utf-8 -*-

from __future__ import absolute_import, print_function, unicode_literals

import sys

import salt.client.ssh
import salt.utils.parsers
from salt.utils.verify import verify_log


class SaltSSH(salt.utils.parsers.SaltSSHOptionParser):
    """
    Used to Execute the salt ssh routine
    """

    def run(self):
        if "-H" in sys.argv or "--hosts" in sys.argv:
            sys.argv += ["x", "x"]  # Hack: pass a mandatory two options
            # that won't be used anyways with -H or --hosts
        self.parse_args()
        self.setup_logfile_logger()
        verify_log(self.config)

        ssh = salt.client.ssh.SSH(self.config)
        ssh.run()



# -*- coding: utf-8 -*-

from __future__ import absolute_import

import logging

from tests.unit.utils.scheduler.base import SchedulerTestsBase

log = logging.getLogger(__name__)


class SchedulerRunJobTest(SchedulerTestsBase):
    """
    Validate the pkg module
    """

    def setUp(self):
        super(SchedulerRunJobTest, self).setUp()
        self.schedule.opts["loop_interval"] = 1

    def test_run_job(self):
        """
        verify that scheduled job runs
        """
        job_name = "test_run_job"
        job = {"schedule": {job_name: {"function": "test.ping"}}}
        # Add the job to the scheduler
        self.schedule.opts.update(job)

        # Run job
        self.schedule.run_job(job_name)
        ret = self.schedule.job_status(job_name)
        expected = {"function": "test.ping", "run": True, "name": "test_run_job"}
        self.assertEqual(ret, expected)



# -*- coding: utf-8 -*-
"""
    :codeauthor: Anthony Shaw <anthonyshaw@apache.org>
"""

# Import Python Libs
from __future__ import absolute_import, print_function, unicode_literals

import salt.modules.libcloud_dns as libcloud_dns

# Import Salt Testing Libs
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase, skipIf


class MockDNSDriver(object):
    def __init__(self):
        pass


def get_mock_driver():
    return MockDNSDriver()


@skipIf(not libcloud_dns.HAS_LIBCLOUD, "No libcloud installed")
class LibcloudDnsModuleTestCase(TestCase, LoaderModuleMockMixin):
    def setup_loader_modules(self):
        module_globals = {
            "_get_driver": MagicMock(return_value=MockDNSDriver()),
            "__salt__": {
                "config.option": MagicMock(
                    return_value={"test": {"driver": "test", "key": "2orgk34kgk34g"}}
                )
            },
        }
        if libcloud_dns.HAS_LIBCLOUD is False:
            module_globals["sys.modules"] = {"libcloud": MagicMock()}

        return {libcloud_dns: module_globals}

    def test_module_creation(self):
        client = libcloud_dns._get_driver("test")
        self.assertFalse(client is None)

    def test_init(self):
        with patch("salt.utils.compat.pack_dunder", return_value=False) as dunder:
            libcloud_dns.__init__(None)
            dunder.assert_called_with("salt.modules.libcloud_dns")



# -*- coding: utf-8 -*-
"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""
# Import Python libs
from __future__ import absolute_import, print_function, unicode_literals

# Import Salt Libs
import salt.states.cmd as cmd
from salt.exceptions import CommandExecutionError

# Import Salt Testing Libs
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class CmdTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.states.cmd
    """

    def setup_loader_modules(self):
        return {cmd: {"__env__": "base"}}

    # 'wait' function tests: 1

    def test_wait(self):
        """
        Test to run the given command only if the watch statement calls it.
        """
        name = "cmd.script"

        ret = {"name": name, "result": True, "changes": {}, "comment": ""}

        self.assertDictEqual(cmd.wait(name), ret)

    # 'wait_script' function tests: 1

    def test_wait_script(self):
        """
        Test to download a script from a remote source and execute it
        only if a watch statement calls it.
        """
        name = "cmd.script"

        ret = {"name": name, "result": True, "changes": {}, "comment": ""}

        self.assertDictEqual(cmd.wait_script(name), ret)

    # 'run' function tests: 1

    def test_run(self):
        """
        Test to run a command if certain circumstances are met.
        """
        name = "cmd.script"

        ret = {"name": name, "result": False, "changes": {}, "comment": ""}

        with patch.dict(cmd.__opts__, {"test": False}):
            comt = "Invalidly-formatted 'env' parameter. See documentation."
            ret.update({"comment": comt})
            self.assertDictEqual(cmd.run(name, env="salt"), ret)

        with patch.dict(cmd.__grains__, {"shell": "shell"}):
            with patch.dict(cmd.__opts__, {"test": False}):
                mock = MagicMock(side_effect=[CommandExecutionError, {"retcode": 1}])
                with patch.dict(cmd.__salt__, {"cmd.run_all": mock}):
                    ret.update({"comment": "", "result": False})
                    self.assertDictEqual(cmd.run(name), ret)

                    ret.update(
                        {
                            "comment": 'Command "cmd.script" run',
                            "result": False,
                            "changes": {"retcode": 1},
                        }
                    )
                    self.assertDictEqual(cmd.run(name), ret)

            with patch.dict(cmd.__opts__, {"test": True}):
                comt = 'Command "cmd.script" would have been executed'
                ret.update({"comment": comt, "result": None, "changes": {}})
                self.assertDictEqual(cmd.run(name), ret)

    def test_run_root(self):
        """
        Test to run a command with a different root
        """
        name = "cmd.script"

        ret = {"name": name, "result": False, "changes": {}, "comment": ""}

        with patch.dict(cmd.__grains__, {"shell": "shell"}):
            with patch.dict(cmd.__opts__, {"test": False}):
                mock = MagicMock(side_effect=[CommandExecutionError, {"retcode": 1}])
                with patch.dict(cmd.__salt__, {"cmd.run_chroot": mock}):
                    ret.update({"comment": "", "result": False})
                    self.assertDictEqual(cmd.run(name, root="/mnt"), ret)

                    ret.update(
                        {
                            "comment": 'Command "cmd.script" run',
                            "result": False,
                            "changes": {"retcode": 1},
                        }
                    )
                    self.assertDictEqual(cmd.run(name, root="/mnt"), ret)

    # 'script' function tests: 1

    def test_script(self):
        """
        Test to download a script and execute it with specified arguments.
        """
        name = "cmd.script"

        ret = {"name": name, "result": False, "changes": {}, "comment": ""}

        with patch.dict(cmd.__opts__, {"test": False}):
            comt = "Invalidly-formatted 'env' parameter. See documentation."
            ret.update({"comment": comt})
            self.assertDictEqual(cmd.script(name, env="salt"), ret)

        with patch.dict(cmd.__grains__, {"shell": "shell"}):
            with patch.dict(cmd.__opts__, {"test": True}):
                comt = "Command 'cmd.script' would have been executed"
                ret.update({"comment": comt, "result": None, "changes": {}})
                self.assertDictEqual(cmd.script(name), ret)

            with patch.dict(cmd.__opts__, {"test": False}):
                mock = MagicMock(side_effect=[CommandExecutionError, {"retcode": 1}])
                with patch.dict(cmd.__salt__, {"cmd.script": mock}):
                    ret.update({"comment": "", "result": False})
                    self.assertDictEqual(cmd.script(name), ret)

                    ret.update(
                        {
                            "comment": "Command 'cmd.script' run",
                            "result": False,
                            "changes": {"retcode": 1},
                        }
                    )
                    self.assertDictEqual(cmd.script(name), ret)

    # 'call' function tests: 1

    def test_call(self):
        """
        Test to invoke a pre-defined Python function with arguments
        specified in the state declaration.
        """
        name = "cmd.script"
        #         func = 'myfunc'

        ret = {"name": name, "result": False, "changes": {}, "comment": ""}

        flag = None

        def func():
            """
            Mock func method
            """
            if flag:
                return {}
            else:
                return []

        with patch.dict(cmd.__grains__, {"shell": "shell"}):
            flag = True
            self.assertDictEqual(cmd.call(name, func), ret)

            flag = False
            ret.update({"comment": "", "result": False, "changes": {"retval": []}})
            self.assertDictEqual(cmd.call(name, func), ret)

    # 'wait_call' function tests: 1

    def test_wait_call(self):
        """
        Test to run wait_call.
        """
        name = "cmd.script"
        func = "myfunc"

        ret = {"name": name, "result": True, "changes": {}, "comment": ""}

        self.assertDictEqual(cmd.wait_call(name, func), ret)

    # 'mod_watch' function tests: 1

    def test_mod_watch(self):
        """
        Test to execute a cmd function based on a watch call
        """
        name = "cmd.script"

        ret = {"name": name, "result": False, "changes": {}, "comment": ""}

        def func():
            """
            Mock func method
            """
            return {}

        with patch.dict(cmd.__grains__, {"shell": "shell"}):
            with patch.dict(cmd.__opts__, {"test": False}):
                mock = MagicMock(return_value={"retcode": 1})
                with patch.dict(cmd.__salt__, {"cmd.run_all": mock}):
                    self.assertDictEqual(
                        cmd.mod_watch(name, sfun="wait", stateful=True), ret
                    )

                    comt = 'Command "cmd.script" run'
                    ret.update({"comment": comt, "changes": {"retcode": 1}})
                    self.assertDictEqual(
                        cmd.mod_watch(name, sfun="wait", stateful=False), ret
                    )

                with patch.dict(cmd.__salt__, {"cmd.script": mock}):
                    ret.update({"comment": "", "changes": {}})
                    self.assertDictEqual(
                        cmd.mod_watch(name, sfun="script", stateful=True), ret
                    )

                    comt = "Command 'cmd.script' run"
                    ret.update({"comment": comt, "changes": {"retcode": 1}})
                    self.assertDictEqual(
                        cmd.mod_watch(name, sfun="script", stateful=False), ret
                    )

                with patch.dict(cmd.__salt__, {"cmd.script": mock}):
                    ret.update({"comment": "", "changes": {}})
                    self.assertDictEqual(
                        cmd.mod_watch(name, sfun="call", func=func), ret
                    )

                    comt = "cmd.call needs a named parameter func"
                    ret.update({"comment": comt})
                    self.assertDictEqual(cmd.mod_watch(name, sfun="call"), ret)

                comt = (
                    "cmd.salt does not work with the watch requisite,"
                    " please use cmd.wait or cmd.wait_script"
                )
                ret.update({"comment": comt})
                self.assertDictEqual(cmd.mod_watch(name, sfun="salt"), ret)



# -*- coding: utf-8 -*-
"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""
# Import Python libs
from __future__ import absolute_import, print_function, unicode_literals

# Import Salt Libs
import salt.states.boto_ec2 as boto_ec2

# Import Salt Testing Libs
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class BotoEc2TestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.states.boto_ec2
    """

    def setup_loader_modules(self):
        return {boto_ec2: {}}

    # 'key_present' function tests: 1

    def test_key_present(self):
        """
        Test to ensure key pair is present.
        """
        name = "mykeypair"
        upublic = "salt://mybase/public_key.pub"

        ret = {"name": name, "result": True, "changes": {}, "comment": ""}

        mock = MagicMock(side_effect=[True, False, False])
        mock_bool = MagicMock(side_effect=[IOError, True])
        with patch.dict(
            boto_ec2.__salt__, {"boto_ec2.get_key": mock, "cp.get_file_str": mock_bool}
        ):
            comt = "The key name {0} already exists".format(name)
            ret.update({"comment": comt})
            self.assertDictEqual(boto_ec2.key_present(name), ret)

            comt = "File {0} not found.".format(upublic)
            ret.update({"comment": comt, "result": False})
            self.assertDictEqual(boto_ec2.key_present(name, upload_public=upublic), ret)

            with patch.dict(boto_ec2.__opts__, {"test": True}):
                comt = "The key {0} is set to be created.".format(name)
                ret.update({"comment": comt, "result": None})
                self.assertDictEqual(
                    boto_ec2.key_present(name, upload_public=upublic), ret
                )

    # 'key_absent' function tests: 1

    def test_key_absent(self):
        """
        Test to deletes a key pair
        """
        name = "new_table"

        ret = {"name": name, "result": True, "changes": {}, "comment": ""}

        mock = MagicMock(side_effect=[False, True])
        with patch.dict(boto_ec2.__salt__, {"boto_ec2.get_key": mock}):
            comt = "The key name {0} does not exist".format(name)
            ret.update({"comment": comt})
            self.assertDictEqual(boto_ec2.key_absent(name), ret)

            with patch.dict(boto_ec2.__opts__, {"test": True}):
                comt = "The key {0} is set to be deleted.".format(name)
                ret.update({"comment": comt, "result": None})
                self.assertDictEqual(boto_ec2.key_absent(name), ret)



# -*- coding: utf-8 -*-
"""
    :codeauthor: Eric Vz <eric@base10.org>
"""

# Import Python Libs
from __future__ import absolute_import

# Import Salt Libs
import salt.modules.pacmanpkg as pacman

# Import Salt Testing Libs
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class PacmanTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.modules.pacman
    """

    def setup_loader_modules(self):
        return {pacman: {}}

    def test_list_pkgs(self):
        """
        Test if it list the packages currently installed in a dict
        """
        cmdmock = MagicMock(return_value="A 1.0\nB 2.0")
        sortmock = MagicMock()
        stringifymock = MagicMock()
        mock_ret = {"A": ["1.0"], "B": ["2.0"]}
        with patch.dict(
            pacman.__salt__,
            {
                "cmd.run": cmdmock,
                "pkg_resource.add_pkg": lambda pkgs, name, version: pkgs.setdefault(
                    name, []
                ).append(version),
                "pkg_resource.sort_pkglist": sortmock,
                "pkg_resource.stringify": stringifymock,
            },
        ):
            self.assertDictEqual(pacman.list_pkgs(), mock_ret)

        sortmock.assert_called_with(mock_ret)
        stringifymock.assert_called_with(mock_ret)

    def test_list_pkgs_as_list(self):
        """
        Test if it lists the packages currently installed in a dict
        """
        cmdmock = MagicMock(return_value="A 1.0\nB 2.0")
        sortmock = MagicMock()
        stringifymock = MagicMock()
        mock_ret = {"A": ["1.0"], "B": ["2.0"]}
        with patch.dict(
            pacman.__salt__,
            {
                "cmd.run": cmdmock,
                "pkg_resource.add_pkg": lambda pkgs, name, version: pkgs.setdefault(
                    name, []
                ).append(version),
                "pkg_resource.sort_pkglist": sortmock,
                "pkg_resource.stringify": stringifymock,
            },
        ):
            self.assertDictEqual(pacman.list_pkgs(True), mock_ret)

        sortmock.assert_called_with(mock_ret)
        self.assertTrue(stringifymock.call_count == 0)

    def test_group_list(self):
        """
        Test if it lists the available groups
        """

        def cmdlist(cmd, **kwargs):
            """
            Handle several different commands being run
            """
            if cmd == ["pacman", "-Sgg"]:
                return "group-a pkg1\ngroup-a pkg2\ngroup-f pkg9\ngroup-c pkg3\ngroup-b pkg4"
            elif cmd == ["pacman", "-Qg"]:
                return "group-a pkg1\ngroup-b pkg4"
            else:
                return "Untested command ({0}, {1})!".format(cmd, kwargs)

        cmdmock = MagicMock(side_effect=cmdlist)

        sortmock = MagicMock()
        with patch.dict(
            pacman.__salt__, {"cmd.run": cmdmock, "pkg_resource.sort_pkglist": sortmock}
        ):
            self.assertDictEqual(
                pacman.group_list(),
                {
                    "available": ["group-c", "group-f"],
                    "installed": ["group-b"],
                    "partially_installed": ["group-a"],
                },
            )

    def test_group_info(self):
        """
        Test if it shows the packages in a group
        """

        def cmdlist(cmd, **kwargs):
            """
            Handle several different commands being run
            """
            if cmd == ["pacman", "-Sgg", "testgroup"]:
                return "testgroup pkg1\ntestgroup pkg2"
            else:
                return "Untested command ({0}, {1})!".format(cmd, kwargs)

        cmdmock = MagicMock(side_effect=cmdlist)

        sortmock = MagicMock()
        with patch.dict(
            pacman.__salt__, {"cmd.run": cmdmock, "pkg_resource.sort_pkglist": sortmock}
        ):
            self.assertEqual(
                pacman.group_info("testgroup")["default"], ["pkg1", "pkg2"]
            )

    def test_group_diff(self):
        """
        Test if it shows the difference between installed and target group contents
        """

        listmock = MagicMock(return_value={"A": ["1.0"], "B": ["2.0"]})
        groupmock = MagicMock(
            return_value={
                "mandatory": [],
                "optional": [],
                "default": ["A", "C"],
                "conditional": [],
            }
        )
        with patch.dict(
            pacman.__salt__, {"pkg.list_pkgs": listmock, "pkg.group_info": groupmock}
        ):
            results = pacman.group_diff("testgroup")
            self.assertEqual(
                results["default"], {"installed": ["A"], "not installed": ["C"]}
            )



# -*- coding: utf-8 -*-
"""
Decorators for salt.state

:codeauthor: :email:`Bo Maryniuk (bo@suse.de)`
"""

# Import Python libs
from __future__ import absolute_import, unicode_literals

import logging

# Import salt libs
import salt.utils.stringutils
from salt.exceptions import SaltException

log = logging.getLogger(__name__)


class OutputUnifier(object):
    def __init__(self, *policies):
        self.policies = []
        for pls in policies:
            if not hasattr(self, pls):
                raise SaltException("Unknown policy: {0}".format(pls))
            else:
                self.policies.append(getattr(self, pls))

    def __call__(self, func):
        def _func(*args, **kwargs):
            result = func(*args, **kwargs)
            for pls in self.policies:
                try:
                    result = pls(result)
                except Exception as exc:  # pylint: disable=broad-except
                    log.debug(
                        "An exception occurred in this state: %s",
                        exc,
                        exc_info_on_loglevel=logging.DEBUG,
                    )
                    result = {
                        "result": False,
                        "name": "later",
                        "changes": {},
                        "comment": "An exception occurred in this state: {0}".format(
                            exc
                        ),
                    }
            return result

        return _func

    def content_check(self, result):
        """
        Checks for specific types in the state output.
        Raises an Exception in case particular rule is broken.

        :param result:
        :return:
        """
        if not isinstance(result, dict):
            err_msg = "Malformed state return. Data must be a dictionary type."
        elif not isinstance(result.get("changes"), dict):
            err_msg = "'Changes' should be a dictionary."
        else:
            missing = []
            for val in ["name", "result", "changes", "comment"]:
                if val not in result:
                    missing.append(val)
            if missing:
                err_msg = "The following keys were not present in the state return: {0}.".format(
                    ", ".join(missing)
                )
            else:
                err_msg = None

        if err_msg:
            raise SaltException(err_msg)

        return result

    def unify(self, result):
        """
        While comments as a list are allowed,
        comments needs to be strings for backward compatibility.
        See such claim here: https://github.com/saltstack/salt/pull/43070

        Rules applied:
          - 'comment' is joined into a multi-line string, in case the value is a list.
          - 'result' should be always either True, False or None.

        :param result:
        :return:
        """
        if isinstance(result.get("comment"), list):
            result["comment"] = "\n".join(
                [salt.utils.stringutils.to_unicode(elm) for elm in result["comment"]]
            )
        if result.get("result") is not None:
            result["result"] = bool(result["result"])

        return result
